import logging
import os
import shutil
from glob import glob
import subprocess
import shlex

from albacore.path_utils import get_default_path

logger = logging.getLogger(__name__)

def find_aligner(config_section, data_path):
    exe = config_section.get('executable', 'ont_bwa')

    # Check we can find executable in the path given
    if shutil.which(exe) is not None:
        logger.info("Found " + exe + " without giving path")
        return shutil.which(exe)

    # If BWA is not installed, or user spec, we try to find it in module path
    # This can be (depending on OS and installation type)
    # either .../data, .../data_versioned or /opt/albacore
    exe_path = data_path if 'data' not in data_path else os.path.dirname(data_path)
    if shutil.which(exe, path=exe_path) is not None:
        logger.info("Found " + exe + " with path " + exe_path)
        return shutil.which(exe, path=exe_path)
    else:
        raise FileNotFoundError('Could not find {} or {}'.format(exe, os.path.join(exe_path, exe)))


def create_index(config_section, reference):
    logger = logging.getLogger('albacore')

    executable = config_section['executable']
    reference_path = config_section['reference_path']
    reference_abs = os.path.abspath(os.path.join(reference_path, reference))

    index_files_template = config_section['index_files'].split()
    index_files = \
        [os.path.abspath(os.path.join(reference_path,
                                      file.format(reference=reference))) \
         for file in index_files_template]
    have_index = True
    for file in index_files:
        if (len(glob(file)) == 0): have_index = False
    have_reference = (len(glob(reference_abs)) > 0)
    if not have_index and not have_reference:
        raise Exception('Could not find reference file {} '
                        'and no index for it either.'.format(reference_abs))
    # one might say that an index should be built whenever there is none
    build_index = not have_index
    # but we also check if the current index is up to date
    if have_reference and have_index:
        index_tms = os.path.getmtime(index_files[0])
        ref_tms = os.path.getmtime(reference_abs)
        build_index = (ref_tms > index_tms)

    if build_index:
        logger.info('Creating index for reference ' + reference_abs)
        index_command = config_section['index_command'].format(executable=executable,
                                                               reference=reference)
        subprocess.Popen(shlex.split(index_command), cwd=reference_path).wait()
        build_success = True
        for file in index_files:
            if (len(glob(file)) == 0): build_success = False
        if not build_success:
            raise Exception('Creation of index for reference ' + reference_abs + ' failed.')


def load_reference(config_section, reference):
    logger = logging.getLogger('albacore')

    executable = config_section['executable']
    reference_path = config_section['reference_path']
    min_len_to_load = int(config_section['preload_threshold'])

    reference_abs = os.path.abspath(os.path.join(reference_path, reference))

    # load index into shared memory
    reference_size = os.stat(reference_abs).st_size
    logger.info('Size of reference file: {}'.format(reference_size))
    if reference_size > min_len_to_load:
        logger.info('Loading {}'.format(reference))
        load_command = config_section['load_command'].format(executable=executable,
                                                             reference=reference)
        ret = subprocess.Popen(shlex.split(load_command), cwd=reference_path).wait()
        logger.info('Return status of index loading: {}'.format(ret))
        return [reference]
    else:
        return []


def initialise_aligner(config, alignref, data_path, section_name):
    logger = logging.getLogger('albacore')

    # use reference from config file if none is given on command line
    if not alignref and config.has_option(section_name, 'reference'):
        alignref = config.get(section_name, 'reference')
    reference_loaded = []

    # disable this module if no either 'method' is empty
    # or 'method' is 'external_process' and no align_command is given
    if config.has_section(section_name):
        if not alignref:
            logger.info('No reference given for module {}'.format(section_name))
        else:
            switched_off = True
            method = config.get(section_name, 'method', fallback=None)
            align_command = config.get(section_name, 'align_command', fallback=None)
            switched_off = not method or method.startswith('#') \
                        or (method == 'external_process' and \
                            (not align_command or align_command.startswith('#')))
            if switched_off:
                logger.info('Module {} is switched off.'.format(section_name))
                alignref = None

    if alignref:  # that is, if alignment is enabled
        if not config.has_section(section_name):
            raise Exception('config section {} not found.'.format(section_name))

        # (1.) get path for references and executable
        if config.has_option(section_name, 'reference_path'):
            config.set(section_name, 'reference_path',
                       config.get(section_name, 'reference_path').\
                           format(data_path=data_path)
                      )
        else:
            config.set(section_name, 'reference_path', '')

        if config.get(section_name, 'method') == 'external_process':
            exe = find_aligner(dict(config.items(section_name)), data_path)
            config.set(section_name, 'executable', exe)
            # (2.) create index
            if config.has_option(section_name, 'index_command'):
                create_index(dict(config.items(section_name)), alignref)

            # (3.) load reference into memory (if large enough)
            if config.has_option(section_name, 'load_command'):
                if not config.has_option(section_name, 'unload_command'):
                    raise Exception(section_name + '.load_command is given, but no ' +
                                    section_name + '.unload_command.')
                reference_loaded = load_reference(dict(config.items(section_name)), alignref)
        if not reference_loaded:
            alignref = os.path.abspath(os.path.join(config.get(section_name, 'reference_path'), alignref))
            if not os.path.exists(alignref):
                raise FileNotFoundError('Could not find alignment reference {}.'.format(alignref))
        config.set(section_name, 'reference', alignref)

        # (4.) remove config options that C++ pipeline does not need
        to_remove = ['reference_path', # aligner.reference is now absolute path
                     'index_command', # indexing and loading has been done
                     'index_files',
                     'load_command',
                     'preload_threshold',
                    ]
        for option in to_remove:
            config.remove_option(section_name, option)
    else:
        config.remove_section(section_name)
        config.add_section(section_name)
        config.set(section_name, 'method', '')
    return reference_loaded
