"""
config_selector.py
Created on: May 18, 2017
Proprietary and confidential information of Oxford Nanopore Technologies, Limited
All rights reserved; (c)2016: Oxford Nanopore Technologies, Limited
"""
from collections import namedtuple
from configparser import ConfigParser
from glob import glob

import logging
import os
import sys

logger = logging.getLogger('albacore')
WORKFLOW_LINEAR = 'linear'
WORKFLOW_1DSQ = '1dsq'


def print_workflows(config_folder, workflow):
    ''' Pretty prints the available flowcell, kit and config combinations
    fir a given workflow type'''
    print('Parsing config files in {}.'.format(config_folder))
    cfg_selector = ConfigSelector(config_folder)
    cfgs = cfg_selector.list_available(workflow)

    format_string = '{0:11} {1:11} {2:10} {3:6}'
    print('Available flowcell + kit combinations are:')
    print(format_string.format('flowcell', 'kit', 'barcoding', 'config file'))
    bcstr = {True: 'included', False: ''}

    for flowcell, kit, config, barcode in sorted(cfgs, key=lambda x: (x[0], x[1])):
        print(format_string.format(flowcell, kit, bcstr[barcode], os.path.basename(config)))
    sys.exit(0)


def choose_config(data_path, flowcell, kit):
    ''' Selects the appropriate config for a given flowcell-kit combination'''
    cfg_selector = ConfigSelector(data_path)
    try:
        config, barcoding, _ = cfg_selector.compatability[(flowcell, kit)]
    except KeyError:
        error_msg = ('Could not find a config file which supports flowcell '
                     '{} and kit {}'.format(flowcell, kit))
        raise KeyError(error_msg)

    logger.info('Selected config file {} to match flowcell {} and '
                'kit {}'.format(config, flowcell, kit))
    return config, barcoding


class ConfigSelector():
    """ This support class facilitates the selection of a configuration
    file based on flowcell and sequencing kit. It also can be used to
    easily decide whether a particular combination is one that includes
    barcoding chemistry. 
    """

    def __init__(self, config_folder):
        """ Constructor.
        @param[in] data_path The path to the configuration files.
        
        This will build up a compatibility table that can be used to select
        the proper configuration file for a particular combination of flowcell
        and sequencing kit. This assumes that there will only be at most ONE
        config file compatible with any particular combination of flowcell and
        sequencing kit.
        """
        configs = glob(os.path.join(config_folder, '*.cfg'))
        compatability = dict()
        for config_path in configs:
            config = ConfigParser(interpolation=None)
            config.read(config_path)
            if 'compatibility' not in config:
                continue
            try:
                if config.has_option('compatibility', '1dsq_config'):
                    workflow = WORKFLOW_1DSQ
                else:
                    workflow = WORKFLOW_LINEAR
                for flowcell in config.get('compatibility', 'compatible_flowcells', fallback='').split(','):
                    if flowcell:
                        for kit in config.get('compatibility', 'compatible_kits', fallback='').split(','):
                            if kit:
                                compatability[(flowcell, kit)] = (config_path, False, workflow)
                        for kit in config.get('compatibility', 'compatible_kits_with_barcoding', fallback='').split(','):
                            if kit:
                                compatability[(flowcell, kit)] = (config_path, True, workflow)
            except Exception:
                continue

        self.compatability = compatability

    def list_available(self, workflow):
        """ Provides a list of valid flowcell+kit combinations for a given
        workflow in the form: (flowcell, kit, config_file, barcoding).
        flowcell, kit are strings, config_file is a path, barcoding is a bool
        """
        return [(flo, kit, conf, barc) for (flo, kit), (conf, barc, work)
                in self.compatability.items() if work == workflow]
