import os
import logging
from configparser import NoSectionError, NoOptionError
from sys import stderr
from configparser import ConfigParser


def resume_options_1d(opts, unknown_args):
    if _resume_options(opts, unknown_args):
        opts.alignref = None
        opts.barcoding = False
        config = ConfigParser(interpolation=None)
        config.read(opts.config)
        opts.barcoding = config.has_option('barcode_detector', 'config_file')


def resume_options_2d(opts, unknown_args):
    _resume_options(opts, unknown_args)


def _resume_options(opts, unknown_args):
    """" Extract the options of the previous run from output directory
       and replace into opts. Sets opts.resume to False if either previous
       log or summary file cannot be found

       :returns: True if previous config file could be found, False otherwise
    """
    # check that there is something to resume
    logfilename = os.path.join(opts.save_path, opts.logfile)
    if not os.path.exists(logfilename):
        print('Could not resume, as logfile {} was not found. '
              'Will process all available reads'.format(logfilename), file=stderr)
        opts.resume = False
        return False

    summaryfilename = os.path.join(opts.save_path, opts.summfile)
    if not os.path.exists(summaryfilename):
        print('Could not resume, as summary file {} was not found. '
              'Will process all available reads'.format(summaryfilename), file=stderr)
        opts.resume = False
        return False

    cfgfilename = os.path.join(opts.save_path, opts.cfgfile)
    if not os.path.exists(cfgfilename):
        print('Could not use previous configuration for resume, '
              'as config file {} was not found. Will continue with '
              'newly passed new config file or flowcell/kit type.'.format(cfgfilename), file=stderr)
        return False
    else:
        opts.config = os.path.join(opts.save_path, opts.cfgfile)
    opts.flowcell = None
    opts.kit = None
    if unknown_args:
        print('Config overrides ({} given) '
              'will be ignored in resume mode.'.format(len(unknown_args)), file=stderr)
        del unknown_args[:]
    return True


def get_barcoding_options(config, barcoding, data_path, logger):
    logger = logging.getLogger('albacore')
    if barcoding:
        if ('barcode_detector' not in config
                or 'config_file' not in config['barcode_detector']):
            msg = ('The "barcode_detector" section or barcoding config location missing '
                   'from config file {}.'.format(config))
            logger.error(msg)
            raise Exception(msg)
        barcoding_config = os.path.join(data_path,
                                        config['barcode_detector']['config_file'])
        config.remove_option('barcode_detector', 'config_file')
        config.read(barcoding_config)
        logger.info('Loaded barcoding config from {}'.format(barcoding_config))
    else:
        try:
            config.remove_option('barcode_detector', 'config_file')
        except (NoSectionError, NoOptionError) as e:
            raise Exception("Missing required config option: barcode_detector.config_file") from e


def find_config_option_else_default(config, section, option, default_value):
    logger = logging.getLogger('albacore')
    if config.has_option(section, option):
        op_value = config.getfloat(section, option)
    else:
        logger.warning("No option '{}.{}' in the config, so using the default"
                       " of {}".format(section, option, default_value))
        op_value = default_value
    return op_value


def parse_config_override(unknown_args):
    """ We accept options of the form --[section].[parameter]=[value]
        and return them as a dict with keys [section].[parameter]

    """
    ret = {}
    for option in unknown_args:
        if not option.startswith('--'):
            raise Exception('Unknown command-line argument ' + option)
        option_tail = option[2:]
        option_parts = option_tail.split('=')
        if len(option_parts) != 2:
            raise Exception('Unknown command-line argument ' + option)
        key = option_parts[0]
        value = option_parts[1]
        split_key = key.split('.')
        if len(split_key) != 2:
            raise Exception('Unknown command-line argument ' + option)
        ret[tuple(split_key)] = value
    return ret


def parse_telemetry_options(config):
    """ Extract telemetry-specific options from a config file.

    :param ConfigParser config: The config object to extract parameters from.
    :returns: tuple of:

        * a list of urls to ping telemetry to
        * the segment duration, in minutes
    :rtype: tuple

    """

    telemetry_urls = []
    telemetry_segment_duration = 60
    telemetry_analysis_id = None
    if 'telemetry' in config:
        telemetry_urls_string = config.get('telemetry', 'urls',
                                           fallback=None)
        if telemetry_urls_string is not None:
            telemetry_urls = telemetry_urls_string.split(',')
        telemetry_segment_duration = config.getint('telemetry',
                                                   'segment_duration',
                                                   fallback=60)
        telemetry_analysis_id = config.get('telemetry',
                                           'analysis_id',
                                           fallback=None)
        config.remove_section('telemetry')
    return telemetry_urls, telemetry_segment_duration, telemetry_analysis_id
