"""
data_handler.py
Created on: November 17, 2016
Proprietary and confidential information of Oxford Nanopore Technologies, Limited
All rights reserved; (c)2016: Oxford Nanopore Technologies, Limited
"""
import logging
import os
from abc import ABC, abstractmethod
from albacore import log_utils


class DataHandler(ABC):
    """ Abstract base class for objects that can be used as data handlers
    by the Pipeline class. Implementations of this base class provide 
    methods for extracting read data from a "data source", and for
    outputting analysis results to wherever they belong. 
    """

    @property    
    @abstractmethod
    def basecall_type(self):
        """ Returns the basecall type, which would normally be specified
        in the workflow configuration. The Pipeline class expects the
        DataHandler object to tell it what kind of basecalling workflow
        it is.
        """
        return None
    
    @abstractmethod
    def load_read_data(self, name):
        """ Load the read data from a data source.
        
        :param name: A string uniquely identifying the read.
        :returns: A tuple containing the data and a dict with read metadata.
        
        The data field will be a dict containing whatever is appropriate
        for the particular data handler implementation.
        
        The returned metadata dict should contain whatever metadata is required
        for the entry point module of the pipeline.
        """
        return None, None
    
    @abstractmethod
    def output_log(self, data_id, data):
        """ Output analysis results from the pipeline.

        :param data_id: The unique identifier for the read.
        :param data: A dict containing at least a key 'log'.
        """

    @abstractmethod
    def output_segmentation(self, data_id, data):
        """ Output analysis results for the segmentation step.
        
        :param data_id: The unique identifier for the read.
        :param data: A dict containing the results of the segmentation
            step of the pipeline analysis.
        :returns: A "tag" object containing information that further "output"
            functions may need.
        """
        return None
    
    @abstractmethod
    def output_basecall_1d(self, data_id, data, seg_tag):
        """ Output analysis results for the 1d basecall step.
        
        :param data_id: The unique identifier for the read.
        :param data: A dict containing the results of the 1d basecall step
            of the pipeline analysis.
        :param seg_tag: The tag returned by the output_segmentation() method.
        :returns: A "tag" object containing information that further "output"
            functions may need.
        """
        return None
    
    @abstractmethod
    def output_basecall_2d(self, data_id, data, seg_tag, bc1d_tag):
        """ Output analysis results for the 2d basecall step.
        
        :param data_id: The unique identifier for the read.
        :param data: A dict containing the results of the 2d basecall step
            of the pipeline analysis.
        :param seg_tag: The tag returned by the output_segmentation() method.
        :param bc1d_tag: The tag returned by the output_basecall_1d() method.
        :returns: A "tag" object containing information that further "output"
            functions may need.
        """
        return None

    @abstractmethod
    def output_barcoding(self, data_id, data, seg_tag, bc1d_tag, bc2d_tag):
        """ Add the Analyses group for the barcoding step.

        :param data_id: The unique identifier for the read.
        :param data: A dict containing the results of the barcoding
            step of the pipeline analysis.
        :param seg_tag: The tag returned by the output_segmentation() method.
        :param bc1d_tag: The tag returned by the output_basecall_1d() method.
        :param bc2d_tag: The tag returned by the output_basecall_2d() method.
        :returns: The group name the barcoding results were written to.
        """
        return None

    @abstractmethod
    def output_alignment(self, data_id, data, seg_tag, bc1d_tag, bc2d_tag):
        """ Add the Analyses group for the alignment step.

        :param data_id: The unique identifier for the read.
        :param data: A dict containing the results of the alignment
            step of the pipeline analysis.
        :param seg_tag: The tag returned by the output_segmentation() method.
        :param bc1d_tag: The tag returned by the output_basecall_1d() method.
        :param bc2d_tag: The tag returned by the output_basecall_2d() method.
        :returns: The group name the alignment results were written to.
        """
        return None

    @abstractmethod
    def output_filtering(self, data_id):
        """
        Output filtering results for the read

        :param data_id: The unique identifier for the read.
        :returns: Boolean of whether the read passes filtering
        """
        return None

    @abstractmethod
    def finish_read(self, data_id):
        """ Notify the Data Handler that no more data will come from this read ID.
        This must function be called by the caller when no functions will be called
        for this data_id.
        :param data_id: The unique identifier for the read.
        """
        return None

    def get_read_subdirectory(self, read_metadata, filter_reads=False,
                              barcoding=False):
        """
        Decide the subdirectory, if any, that the read should be written in to
        (before batching is taken into account). This is used in finish_read
        by all the data handlers.
        :param read_metadata: ReadMetadata instance for this read_id
        :param filter_reads: Boolean of whether filtering reads is enabled
        :param barcoding: Boolean of whther barcoding is enabled
        :return: String subdirectory for this read_id
        """
        filter_dir = ''
        if filter_reads:
            if read_metadata.read_is_calib():
                filter_dir = 'calibration_strands'
                barcoding = False
            elif read_metadata.read_passes_filter():
                filter_dir = 'pass'
            else:
                filter_dir = 'fail'

        barcode_dir = ''
        if barcoding:
            barcode_dir = read_metadata.barcode_arrangement

        sub_dir = os.path.join(filter_dir, barcode_dir)

        return sub_dir

    def _init_batch_counter(self, opts):
        """ Initialises variables self.current_batch_counter and self.current_in_batch_counter.
            
            self.current_in_batch_counter contains a dict from subdirectory to int.
            The subdirectory is the filtering directory, if applicabale (i.e. pass
            or fail), and then the barcode, if applicable.
            With each new read, self.current_in_batch_counter[sub_dir] will be
            increase in self._increment_in_batch_counter() until it reaches self.batch_size.
            Then self.current_batch_counter[sub_dir] is incremented and
            self.current_in_batch_counter reset to 0.
            If filtering is disabled and barcode subfolders are not used, all
            reads will use the key/subdirectory ''.
            
            Assures that in resume mode, batches of the previous and of the new run
            are distinct (that means, half-filled batches from the previous run will
            NOT be filled up).
        """
        if opts.resume and self.batch_size > 0:
            logger = logging.getLogger('albacore')
            file_type = self.output_format
            logfile = open(os.path.join(opts.save_path, opts.logfile), 'r')
            self.current_batch_counter, self.current_in_batch_counter = \
                log_utils.find_batches(file_type, logfile.readlines())          
            for sub_dir in self.current_batch_counter:
                batch_num = self.current_batch_counter[sub_dir]
                logger.info(log_utils.log_resume_batch(file_type, batch_num, sub_dir))
        else:
            self.current_batch_counter = {}
            self.current_in_batch_counter = {}

    def _increment_in_batch_counter(self, sub_dir):
        """ Increases self.current_in_batch_counter[sub_dir] until it reaches self.batch_size.
            Then self.current_batch_counter[sub_dir] is incremented and
            self.current_in_batch_counter reset to 0.
            :returns: pair of ints, containing batch_counter and in_batch_counter
        """
        logger = logging.getLogger('albacore')
        file_type = self.output_format
        # uses the directory as the key
        if self.batch_size > 0:
            if sub_dir not in self.current_batch_counter:
                self.current_in_batch_counter[sub_dir] = -1
                self.current_batch_counter[sub_dir] = 0
                logger.info(log_utils.log_open_batch(file_type,
                                                     self.current_batch_counter[sub_dir],
                                                     sub_dir))
            elif self.current_in_batch_counter[sub_dir] == self.batch_size - 1:
                self.current_batch_counter[sub_dir] += 1
                self.current_in_batch_counter[sub_dir] = -1
                logger.info(log_utils.log_open_batch(file_type,
                                                     self.current_batch_counter[sub_dir],
                                                     sub_dir))
            self.current_in_batch_counter[sub_dir] += 1
            in_batch_counter = self.current_in_batch_counter[sub_dir]
            batch_counter = self.current_batch_counter[sub_dir]
        else:
            batch_counter = -1
            in_batch_counter = -1
        return batch_counter, in_batch_counter
