"""
fast5_data_handler.py
Created on: November 14, 2016
Proprietary and confidential information of Oxford Nanopore Technologies, Limited
All rights reserved; (c)2016: Oxford Nanopore Technologies, Limited
"""
import os
import shutil
import logging
import numpy as np
from datetime import datetime
from configparser import ConfigParser
from albacore.data_handler import DataHandler
from albacore import __version__ as VERSION
from albacore.read_metadata import ReadMetadata
from ont_fast5_api.fast5_file import Fast5File
from ont_fast5_api.analysis_tools.basecall_1d import Basecall1DTools
from ont_fast5_api.analysis_tools.basecall_2d import Basecall2DTools
import albacore.return_statuses as return_statuses
import albacore.config_utils as config_utils

QSCORE_DEFAULT = 0.0


class Fast5DataHandler(DataHandler):
    """ A DataHandler subclass for working with read fast5 files.
    """

    def __init__(self, config_file, opts):
        """ Constructs the object.

        :param config_file: A configuration file with the basecall configuration
            details used for basecalling.
        """
        logger = logging.getLogger('albacore')
        self.config = ConfigParser(interpolation=None)
        self.config.read(config_file)
        self.workflow_type = self.config.get('pipeline', 'basecall_type')
        self.output_format = 'fast5'

        has_barcoding_module = self.config.has_option('barcode_detector',
                                                      'arrangements_files')
        self.make_barcoding_subfolders = self.config.getboolean(
            'barcode_detector',
            'make_subfolders',
            fallback=has_barcoding_module)
        self.data_id_mapping = {}
        self.batch_size = opts.files_per_batch_folder
        self._init_batch_counter(opts)

        self.filter_reads = True
        if opts.disable_filtering is True:
            self.filter_reads = False

        self.minqscore_1d = config_utils.find_config_option_else_default(
            self.config, 'pipeline', 'min_qscore_1d', QSCORE_DEFAULT)
        self.minqscore_1dsq = config_utils.find_config_option_else_default(
            self.config, 'pipeline', 'min_qscore_1dsq', QSCORE_DEFAULT)


        if self.config.has_option('fast5', 'basecall_columns'):
            self.basecall_columns = self.config.get('fast5', 'basecall_columns')
            if len(self.basecall_columns) == 0:
                logger.warning('Empty basecall_columns entry found.')
                logger.warning('Resetting to "all".')
                self.basecall_columns = 'all'
        else:
            self.basecall_columns = 'all'
        self.basecall_columns = self.basecall_columns.split(',')

    @property
    def basecall_type(self):
        """ Returns the basecall type, as specified in the configuration provided
        to the constructor.
        """
        return self.workflow_type

    def load_read_data(self, fname):
        """ Load the read data from the fast5 file fname.
        
        :param fname: A read fast5 file containing raw data.
        :returns: A tuple containing the scaled raw data and a ReadMetadata
            object containing the required metadata.
        """
        read_data = ReadMetadata(fname)
        self.data_id_mapping[read_data.data_id] = read_data
        return {"raw": read_data.raw}, read_data

    def output_log(self, data_id, data):
        """ Output analysis results from the pipeline.

        :param data_id: The unique identifier for the read.
        :param data: A dict containing at least a key 'log'.
        """
        if 'log' not in data: return

        read_info = self.data_id_mapping[data_id]
        fname = read_info.filename
        with Fast5File(fname, 'r+') as fh:
            log_group = fh.add_log('Albacore_Log', datetime.today().strftime('%Y-%b-%d_%H:%M:%S'), data['log'])

    def output_segmentation(self, data_id, data):
        """ Add the Analyses group for the segmentation step.
        
        :param data_id: The unique identifier for the read.
        :param data: A dict containing the results of the segmentation
            step of the pipeline analysis.
        :returns: The group name the segmentation results were written to.
        """
        read_info = self.data_id_mapping[data_id]
        fname = read_info.filename
        with Fast5File(fname, 'r+') as fh:
            segmentation_group = fh.get_latest_analysis('Segmentation', increment=True)
            meta = {'name': 'ONT Albacore Sequencing Software',
                    'version': VERSION,
                    'time_stamp': datetime.today().strftime('%Y-%b-%d %H:%M:%S')}
            config_fields = ['data_trimmer']
            if self.config.has_section('event_detector'):
                config_fields += ['event_detector']
            if self.workflow_type in ['template_complement', 'full_2d']:
                config_fields.append('hairpin_splitter')
            fast5_section_names = {'data_trimmer': 'stall_removal',
                                   'hairpin_splitter': 'split_hairpin',
                                   'event_detector': 'event_detection'}
            config_data = {}
            for section in config_fields:
                config_data[fast5_section_names[section]] = dict(self.config.items(section))
            fh.add_analysis('segmentation', segmentation_group, meta, config_data)

            summary = {'has_template': 0,
                       'has_complement': 0}
            return_status = return_statuses.NO_DATA_FOUND
            if data:
                summary['has_template'] = 1 if data.get('has_template') else 0
                summary['has_complement'] = 1 if data.get('has_complement') else 0
                if summary['has_template'] == 1:
                    summary.update({'num_events_template': data['num_events_template'],
                                    'first_sample_template': data['first_sample_template'],
                                    'duration_template': data['duration_template']})
                if summary['has_complement'] == 1:
                    summary.update({'num_events_complement': data['num_events_complement'],
                                    'first_sample_complement': data['first_sample_complement'],
                                    'duration_complement': data['duration_complement']})
            fh.set_summary_data(segmentation_group, 'segmentation', summary)

            # Determine the return status.
            if data:
                if self.workflow_type == 'linear':
                    return_status = return_statuses.WORKFLOW_SUCCESSFUL if data.get('has_template') else return_statuses.NO_TEMPLATE_DATA
                else:
                    if data.get('has_template'):
                        return_status = return_statuses.WORKFLOW_SUCCESSFUL if data.get('has_complement') else return_statuses.NO_COMPLEMENT_DATA
                    else:
                        return_status = return_statuses.NO_TEMPLATE_DATA
            else:
                return_status = return_statuses.NO_TEMPLATE_DATA
            fh.add_analysis_attributes('{}/Summary'.format(segmentation_group), {'return_status': return_status})
        return segmentation_group

    def output_basecall_1d(self, data_id, data, seg_group):
        """ Add the Analyses group for the 1D basecall step.
        
        :param data_id: The unique identifier for the read.
        :param data: A dict containing the results of the 1D basecall
            step of the pipeline analysis.
        :param seg_group: The group name the segmentation results were written to.
        :returns: The group name the 1D basecall results were written to.
        """
        read_info = self.data_id_mapping[data_id]
        fname = read_info.filename
        template_dataset = data.get('template_data')
        complement_dataset = data.get('complement_data')
        template_fastq = data.get('template_fastq')
        complement_fastq = data.get('complement_fastq')

        with Fast5File(fname, 'r+') as fh:
            basecall_1d_group = fh.get_latest_analysis('Basecall_1D', increment=True)
            meta = {'name': 'ONT Albacore Sequencing Software',
                    'version': VERSION,
                    'time_stamp': datetime.today().strftime('%Y-%b-%d %H:%M:%S'),
                    'segmentation': "Analyses/" + seg_group}
            if self.config.has_section('basecaller'):
                config_fields = ['basecaller']
            elif self.config.has_section('basecaller_hmm'):
                config_fields = ['basecaller_hmm']
            elif self.config.has_section('basecaller_ocl'):
                config_fields = ['basecaller_ocl']
            else:
                raise Exception('Config file has no recognized basecalling section.')
            fast5_section_names = {'basecaller': 'basecall_1d',
                                   'basecaller_ocl': 'basecall_1d',
                                   'basecaller_hmm': 'basecall_1d'}
            config_data = {}
            for section in config_fields:
                config_data[fast5_section_names[section]] = dict(self.config.items(section))
            with Basecall1DTools(fh, group_name=basecall_1d_group, meta=meta, config=config_data) as bc:
                has_template = False
                has_complement = False
                sections = []
                if data:
                    has_template = ('num_events_template' in data and data['num_events_template'] > 0)
                    has_complement = ('num_events_complement' in data and data['num_events_complement'] > 0)
                    if has_template:
                        sections.append('template')
                        # see if the read passes the qscore filter
                        passes_filter = False
                        if data['mean_qscore_template'] >= self.minqscore_1d:
                            passes_filter = True
                        read_info.add_read_filter_status(passes_filter)
                    if has_complement:
                        sections.append('complement')

                for section in sections:
                    event_data = data['{}_data'.format(section)]
                    fastq = data['{}_fastq'.format(section)]
                    if self.basecall_columns == ['all']:
                        bc.add_event_data(section, event_data)
                    else:
                        event_subset = event_data[self.basecall_columns]
                        if len(event_subset) == 0:
                            logger = logging.getLogger('albacore')
                            logger.error('Event table column selection '
                                         'removed all data.')
                            logger.error('Writing all columns.')
                            event_subset = event_data
                        try:
                            bc.add_event_data(section, event_subset)
                        except:
                            # Usually this means all columns were removed.
                            logger = logging.getLogger('albacore')
                            logger.error('Invalid event table column selection:'
                                         ' {}'.format(self.basecall_columns))
                            logger.error('Writing all columns.')
                            bc.add_event_data(section, event_data)
                    bc.add_called_sequence(section, fastq['name'], fastq['sequence'], fastq['qstring'])
                    summary = {'num_events': data['num_events_{}'.format(section)],
                               'called_events': data['num_called_{}'.format(section)],
                               'mean_qscore': data['mean_qscore_{}'.format(section)],
                               'strand_score': data['strand_score_{}'.format(section)],
                               'sequence_length': data['sequence_length_{}'.format(section)],
                               'stay_prob': data['stay_prob_{}'.format(section)],
                               'step_prob': data['step_prob_{}'.format(section)],
                               'skip_prob': data['skip_prob_{}'.format(section)]}
                    fh.set_summary_data(basecall_1d_group, 'basecall_1d_{}'.format(section), summary)

            # Determine the return status.
            if self.workflow_type == 'linear':
                if has_template:
                    if passes_filter:
                        return_status = return_statuses.WORKFLOW_SUCCESSFUL
                    else:
                        return_status = return_statuses.BASECALL_FILTER_FAILURE_1D
                else:
                    return_status = return_statuses.NO_TEMPLATE_DATA
            else:
                if has_template:
                    if has_complement:
                        return_status = return_statuses.WORKFLOW_SUCCESSFUL
                    else:
                        return_status = return_statuses.NO_COMPLEMENT_DATA
                else:
                    return_status = return_statuses.NO_TEMPLATE_DATA
            fh.add_analysis_attributes('{}/Summary'.format(basecall_1d_group),
                                       {'return_status': return_status})
        return basecall_1d_group
    
    def output_basecall_2d(self, data_id, data, seg_group, bc1d_group):
        """ Add the Analyses group for the 2D basecall step.
        
        :param data_id: The unique identifier for the read.
        :param data: A dict containing the results of the 2D basecall
            step of the pipeline analysis.
        :param seg_group: The group name the segmentation results were written to.
        :param bc1d_group: The group name the 1D basecall results were written to.
        :returns: The group name the 2D basecall results were written to.
        """
        read_info = self.data_id_mapping[data_id]
        fname = read_info.filename
        alignment = data.get('2d_alignment_data')
        labls = data.get('2d_labl_data')
        fastq = data.get('2d_fastq')

        with Fast5File(fname, 'r+') as fh:
            basecall_2d_group = fh.get_latest_analysis('Basecall_2D', increment=True)
            meta = {'name': 'ONT Albacore Sequencing Software',
                    'version': VERSION,
                    'time_stamp': datetime.today().strftime('%Y-%b-%d %H:%M:%S'),
                    'segmentation': "Analyses/" + seg_group,
                    'basecall_1d': "Analyses/" + bc1d_group}

            config_fields = ['basecaller_2d']
            config_data = {}
            for section in config_fields:
                config_data[section] = dict(self.config.items(section))

            with Basecall2DTools(fh, group_name=basecall_2d_group, meta=meta, config=config_data) as bc:
                return_status = return_statuses.BASECALL_FAILURE_2D
                if data and alignment is not None and len(alignment) > 0:
                    bc.add_called_sequence("2D", fastq['name'], fastq['sequence'], fastq['qstring'])
                    labllen = len(labls[0])
                    labl_dtype = 'S{}'.format(labllen)
                    align_data = np.fromiter(zip(alignment[0],
                                                 alignment[1],
                                                 labls),
                                             dtype=[('template', np.int32),
                                                    ('complement', np.int32),
                                                    ('model_state', labl_dtype)],
                                             count=alignment[0].size)
                    bc.add_2d_call_alignment(align_data)
                    summary = {'mean_qscore': data['mean_qscore_2d'],
                               'sequence_length': data['sequence_length_2d']}
                    fh.set_summary_data(basecall_2d_group, 'basecall_2d', summary)

                    passes_filter = False
                    if data['mean_qscore_2d'] >= self.minqscore_1dsq:
                        passes_filter = True
                    self.data_id_mapping[read_id].add_read_filter_status(
                        passes_filter)
                    if passes_filter:
                        return_status = return_statuses.WORKFLOW_SUCCESSFUL
                    else:
                        return_status = return_statuses.BASECALL_FILTER_FAILURE_2D
            fh.add_analysis_attributes('{}/Summary'.format(basecall_2d_group), {'return_status': return_status})
        return basecall_2d_group

    def output_barcoding(self, data_id, data, seg_group, bc1d_group, bc2d_group):
        """ Add the Analyses group for the barcoding step.

        :param data_id: The unique identifier for the read.
        :param data: A dict containing the results of the barcoding
            step of the pipeline analysis.
        :param seg_group: The group name the segmentation results were written to.
        :param bc1d_group: The group name the 1D basecall results were written to.
        :param bc2d_group: The group name the 2D basecall results were written to (if any).
        :returns: The group name the barcoding results were written to.
        """
        read_info = self.data_id_mapping[data_id]
        fname = read_info.filename

        barcode_arr = data.get('barcode_arrangement', '')
        if not barcode_arr: return

        with Fast5File(fname, 'r+') as fh:
            barcoding_group = fh.get_latest_analysis('Barcoding', increment=True)
            meta = {'name': 'ONT Albacore Sequencing Software',
                    'version': VERSION,
                    'time_stamp': datetime.today().strftime('%Y-%b-%d %H:%M:%S'),
                    'segmentation': "Analyses/" + seg_group,
                    'basecall_1d': "Analyses/" + bc1d_group}
            if bc2d_group is not None:
                meta['basecall_2d'] = "Analyses/" + bc2d_group

            config_fields = ['barcode_detector']
            config_data = {}
            for section in config_fields:
                config_data[section] = dict(self.config.items(section))
            fh.add_analysis('barcoding', barcoding_group, meta, config_data)

            read_info.add_barcode_arrangement(barcode_arr)
            barcode_found = len(barcode_arr) > 0
            return_status = return_statuses.WORKFLOW_SUCCESSFUL if barcode_found else return_statuses.BARCODE_CLASSIFICATION_FAILED
            fh.add_analysis_attributes('{}/Summary'.format(barcoding_group), {'return_status': return_status})

            if data:
                del data['read_id']
                fh.set_summary_data(barcoding_group, 'barcoding', data)

        return barcoding_group

    def _output_align_result(self, data_id, data, seg_group, bc1d_group, bc2d_group, group_name, config_name, prefix):
        """ Add the Analyses group for the alignment step.

        :param data_id: The unique identifier for the read.
        :param data: A dict containing the results of the alignment
            step of the pipeline analysis.
        :param seg_group: The group name the segmentation results were written to.
        :param bc1d_group: The group name the 1D basecall results were written to.
        :param bc2d_group: The group name the 2D basecall results were written to (if any).
        :returns: The group name the alignment results were written to.
        """
        read_info = self.data_id_mapping[data_id]
        fname = read_info.filename
        prefix_ = prefix + '_'

        with Fast5File(fname, 'r+') as fh:
            alignment_group = fh.get_latest_analysis(group_name, increment=True)
            meta = {'name': 'ONT Albacore Sequencing Software',
                    'version': VERSION,
                    'time_stamp': datetime.today().strftime('%Y-%b-%d %H:%M:%S'),
                    'segmentation': "Analyses/" + seg_group,
                    'basecall_1d': "Analyses/" + bc1d_group}
            if bc2d_group is not None:
                meta['basecall_2d'] = "Analyses/" + bc2d_group

            config_fields = [config_name]
            config_data = {}
            for section in config_fields:
                config_data[section] = dict(self.config.items(section))
            fh.add_analysis(prefix, alignment_group, meta, config_data)

            for section, section_data in data.items():
                del section_data['data_id']
                del section_data['channel']
                # remove 'alignment_' at begin of each key
                section_data_copy = {}
                for key, value in section_data.items():
                    if key.startswith(prefix_):
                        newkey = key[len(prefix_):]
                    else:
                        newkey = key
                    section_data_copy[newkey] = value
                sam_output = None
                if 'sam_output' in section_data_copy:
                    sam_output = section_data_copy.pop('sam_output')
                    fh.add_analysis_subgroup(alignment_group, 'Aligned_{}'.format(section))
                    group_name = '{}/Aligned_{}'.format(alignment_group, section)
                    fh.add_analysis_dataset(group_name, 'SAM', sam_output)
                    return_status = 'alignment successful'
                else:
                    return_status = 'alignment failed'
                fh.set_summary_data(alignment_group, '{}{}'.format(prefix_, section), section_data_copy)
                fh.add_analysis_attributes('{}/Summary'.format(alignment_group), {'return_status': return_status})
                if sam_output is not None:
                    section_data_copy['sam_output'] = sam_output

        return alignment_group

    def output_alignment(self, data_id, data, seg_group, bc1d_group, bc2d_group):
        return self._output_align_result(data_id, data, seg_group, bc1d_group, bc2d_group, \
                                         'Alignment', 'aligner', 'alignment')


    def output_calib_detection(self, data_id, data, seg_group, bc1d_group, bc2d_group):
        for section, section_data in data.items():
            if len(section_data.get('calibration_strand_sam_output', '')) > 0:
                read_info = self.data_id_mapping[data_id]
                read_info.add_read_calib_status(True)
        return self._output_align_result(data_id, data, seg_group, bc1d_group, bc2d_group, \
                                         'Calibration_Strand_Detection', 'calib_detector', 'calibration_strand')

    def output_filtering(self, data_id):
        """
        Find whether a read passed or failed filtering
        :param data_id: The unique identifier for the read.
        :return: Boolean of whether the read passes or failed filtering
        """
        return self.data_id_mapping[data_id].read_passes_filter()

    def _move_to_subfolder(self, data_id):
        """ Moves a read file into filtering, barcoding and/or batch subfolder
        Example: If the read file name is currently /your/folder/aaa.fast5,
        the barcode 03 has been detected by raw2call, and both our options are
        True, it will be copied to /your/folder/barcode03/0/aaa.fast5, where
        "0" increments to "1" after self.batch_size many reads.
        If filtering is also enabled it will go to
        /your/folder/pass/barcode03/0/aaa.fast5

        :param read_id: The unique identifier of the read to be copied
        """
        read_info = self.data_id_mapping[data_id]
        old_filename = read_info.filename
        parent_dir = os.path.dirname(old_filename)

        sub_dir = self.get_read_subdirectory(read_info, self.filter_reads,
                                             self.make_barcoding_subfolders)

        batch_counter, _ = self._increment_in_batch_counter(sub_dir)
        if batch_counter >= 0:
            sub_dir = os.path.join(sub_dir, str(batch_counter))

        target_dir = os.path.join(parent_dir, sub_dir)

        # ensure that both paths end in '/'
        read_dir = os.path.join(parent_dir, '')
        target_dir = os.path.join(target_dir, '')

        new_filename = old_filename.replace(read_dir, target_dir)
        self._move_file(old_filename, new_filename, target_dir, data_id)

    def _move_file(self, old_filename, new_filename, target_dir, data_id):
        try:
            if not os.path.exists(target_dir):
                os.makedirs(target_dir)
            shutil.move(old_filename, new_filename)
            self.data_id_mapping[data_id].replace_filename(new_filename)
        except Exception:
            logger = logging.getLogger('albacore')
            logger.warning('Could not copy file {} into folder {}. '
                           'Will remain in previous place.'.format(old_filename, target_dir))

    def finish_read(self, data_id):
        """ Notify the Data Handler that no more data will come from this read ID.
        This must function be called by the caller when no functions will be called
        for this read_id.
        :param read_id: The unique identifier for the read.
        """
        if self.make_barcoding_subfolders or self.batch_size > 0 \
                or self.filter_reads:
            self._move_to_subfolder(data_id)
        del self.data_id_mapping[data_id]
