"""
fast5_fastq_data_handler.py
Created on: April 26, 2017
Proprietary and confidential information of Oxford Nanopore Technologies, Limited
All rights reserved; (c)2017: Oxford Nanopore Technologies, Limited
"""
import os
import shutil
import numpy as np
from datetime import datetime
from configparser import ConfigParser
from albacore.data_handler import DataHandler
from albacore import __version__ as VERSION
from albacore.fast5_data_handler import Fast5DataHandler
from albacore.fastq_data_handler import FastQDataHandler


class Fast5FastQDataHandler(DataHandler):
    """ A DataHandler subclass for loading Fast5 and writing to Fast5 and/or FastQ files.
    Input always comes from Fast5 file. Output functions call corresponding functions
    of Fast5DataHandler if the configuration parameter pipeline.output contains 'fast5'
    and corresponding function of FastQDataHandler if the configuration parameter
    pipeline.output contains 'fastq'.
    """

    def __init__(self, config_file, opts):
        """ Constructs the object.
        
        :param config_file: A configuration file with the basecall configuration
            details used for basecalling.
        """
        self.fast5_handler = Fast5DataHandler(config_file, opts)
        self.output_format = opts.output_format
        self.write_fast5 = ('fast5' in self.output_format)
        self.write_fastq = ('fastq' in self.output_format)
        if not self.write_fast5 and not self.write_fastq:
            raise Exception('No output method specified '
                            '(allowed values: fast5, fastq and all string containing those).')
        if self.write_fastq:
            self.fastq_handler = FastQDataHandler(config_file, opts)

    @property
    def basecall_type(self):
        """ Returns the basecall type, as specified in the configuration provided
        to the constructor.
        """
        return self.fast5_handler.basecall_type

    def load_read_data(self, fname):
        """ Load the read data from the fast5 file fname.
        
        :param fname: A read fast5 file containing raw data.
        :returns: A tuple containing the scaled raw data and ReadMetadata object
            (defined in fast5_data_handler) with the required metadata.
        """
        data, metadata = self.fast5_handler.load_read_data(fname)
        if self.write_fastq:
            self.fastq_handler.set_read_data(metadata)
        return data, metadata

    def output_log(self, data_id, data):
        """ Output analysis results from the pipeline.

        :param data_id: The unique identifier for the read.
        :param data: A dict containing at least a key 'log'.
        """
        if self.write_fastq:
            self.fastq_handler.output_log(data_id, data)
        if self.write_fast5:
            self.fast5_handler.output_log(data_id, data)

    def output_segmentation(self, data_id, data):
        """ Add the Analyses group for the segmentation step.
        
        :param data_id: The unique identifier for the read.
        :param data: A dict containing the results of the segmentation
            step of the pipeline analysis.
        :returns: The group name the segmentation results were written to.
        """
        if self.write_fastq:
            self.fastq_handler.output_segmentation(data_id, data)
        if self.write_fast5:
            return self.fast5_handler.output_segmentation(data_id, data)
        else:
            return None

    def output_basecall_1d(self, data_id, data, seg_group):
        """ Add the Analyses group for the 1D basecall step.
        
        :param data_id: The unique identifier for the read.
        :param data: A dict containing the results of the 1D basecall
            step of the pipeline analysis.
        :param seg_group: The group name the segmentation results were written to.
        :returns: The group name the 1D basecall results were written to.
        """
        if self.write_fastq:
            self.fastq_handler.output_basecall_1d(data_id, data, seg_group)
        if self.write_fast5:
            return self.fast5_handler.output_basecall_1d(data_id, data, seg_group)
        else:
            return None
    
    def output_basecall_2d(self, data_id, data, seg_group, bc1d_group):
        """ Add the Analyses group for the 2D basecall step.
        
        :param data_id: The unique identifier for the read.
        :param data: A dict containing the results of the 2D basecall
            step of the pipeline analysis.
        :param seg_group: The group name the segmentation results were written to.
        :param bc1d_group: The group name the 1D basecall results were written to.
        :returns: The group name the 2D basecall results were written to.
        """
        if self.write_fastq:
            self.fastq_handler.output_basecall_2d(data_id, data, seg_group, bc1d_group)
        if self.write_fast5:
            return self.fast5_handler.output_basecall_2d(data_id, data, seg_group, bc1d_group)
        else:
            return None

    def output_calib_detection(self, data_id, data, seg_group, bc1d_group, bc2d_group):
        """ Add the Analyses group for the calibration strand detection step.

        :param data_id: The unique identifier for the read.
        :param data: A dict containing the results of the calibration strand detection
            step of the pipeline analysis.
        :param seg_group: The group name the segmentation results were written to.
        :param bc1d_group: The group name the 1D basecall results were written to.
        :param bc2d_group: The group name the 2D basecall results were written to (if any).
        :returns: The group name the alignment results were written to.
        """
        if self.write_fastq:
            self.fastq_handler.output_calib_detection(data_id, data, seg_group, bc1d_group, bc2d_group)
        if self.write_fast5:
            return self.fast5_handler.output_calib_detection(data_id, data, seg_group, bc1d_group, bc2d_group)
        else:
            return None

    def output_barcoding(self, data_id, data, seg_group, bc1d_group, bc2d_group):
        """ Add the Analyses group for the barcoding step.

        :param data_id: The unique identifier for the read.
        :param data: A dict containing the results of the barcoding
            step of the pipeline analysis.
        :param seg_group: The group name the segmentation results were written to.
        :param bc1d_group: The group name the 1D basecall results were written to.
        :param bc2d_group: The group name the 2D basecall results were written to (if any).
        :returns: The group name the barcoding results were written to.
        """
        if self.write_fastq:
            self.fastq_handler.output_barcoding(data_id, data, seg_group, bc1d_group, bc2d_group)
        if self.write_fast5:
            return self.fast5_handler.output_barcoding(data_id, data, seg_group, bc1d_group, bc2d_group)
        else:
            return None

    def output_alignment(self, data_id, data, seg_group, bc1d_group, bc2d_group):
        """ Add the Analyses group for the alignment step.

        :param data_id: The unique identifier for the read.
        :param data: A dict containing the results of the alignment
            step of the pipeline analysis.
        :param seg_group: The group name the segmentation results were written to.
        :param bc1d_group: The group name the 1D basecall results were written to.
        :param bc2d_group: The group name the 2D basecall results were written to (if any).
        :returns: The group name the alignment results were written to.
        """
        if self.write_fastq:
            self.fastq_handler.output_alignment(data_id, data, seg_group, bc1d_group, bc2d_group)
        if self.write_fast5:
            return self.fast5_handler.output_alignment(data_id, data, seg_group, bc1d_group, bc2d_group)
        else:
            return None

    def output_filtering(self, data_id):
        """
        Find whether a read passed or failed filtering
        :param data_id: The unique identifier for the read.
        :return: Boolean of whether the read passes or failed filtering
        """
        if self.write_fastq:
            return self.fastq_handler.output_filtering(data_id)
        if self.write_fast5:
            return self.fast5_handler.output_filtering(data_id)

    def finish_read(self, data_id):
        """ Notify the Data Handler that no more data will come from this read ID.
        This function is unnecessary for Fast5, but needed for FastQ data.
        :param data_id: The unique identifier for the read.
        """
        if self.write_fastq:
            self.fastq_handler.finish_read(data_id)
        if self.write_fast5:
            self.fast5_handler.finish_read(data_id)
        else:
            del self.fast5_handler.data_id_mapping[data_id]
