"""
fastq_data_handler.py
Created on: April 25, 2017
Proprietary and confidential information of Oxford Nanopore Technologies, Limited
All rights reserved; (c)2017: Oxford Nanopore Technologies, Limited
"""
import os
import logging
from configparser import ConfigParser
from albacore.data_handler import DataHandler
import albacore.path_utils as path_utils
import albacore.config_utils as config_utils

QSCORE_DEFAULT = 0.0


class FastQDataHandler(DataHandler):
    """ A DataHandler subclass for writing FastQ files.
    Note that fastQ is only an output format. This class has no implementation of load functions.
    """

    ###################################### INTERFACE METHODS ######################################

    def __init__(self, config_file, opts):
        """ Constructs the object.
        
        :param config_file: A configuration file with the basecall configuration
            details used for basecalling.
        """
        logger = logging.getLogger('albacore')
        self.config = ConfigParser(interpolation=None)
        self.config.read(config_file)
        self.workflow_type = self.config.get('pipeline', 'basecall_type')
        self.output_format = 'fastq'

        self.data = {}  # {read_id : ReadMetatdata()}

        self.barcoding = opts.barcoding
        self.make_barcoding_subfolders = self.config.getboolean(
            'barcode_detector',
            'make_subfolders',
            fallback=self.barcoding)
        self.batch_size = opts.reads_per_fastq_batch
        self._init_batch_counter(opts)

        header_without_barcoding_template = self.config.get('fastq', 'header')
        header_with_barcoding_template = self.config.get('fastq',
                                                         'header_with_barcoding')
        self.header_template = \
            header_with_barcoding_template if self.barcoding \
                else header_without_barcoding_template
        batch_file_name_template = self.config.get('fastq', 'batch_file_name')
        single_file_name_template = self.config.get('fastq', 'single_file_name')
        all_file_name_template = self.config.get('fastq', 'all_file_name')

        if self.batch_size > 1:
            self.file_name_template = batch_file_name_template
        elif self.batch_size == 1:
            self.file_name_template = single_file_name_template
        elif self.batch_size == 0:
            self.file_name_template = all_file_name_template
        else:
            raise ValueError("Cannot choose a negative number for reads per "
                             "fastq batch: {}".format(self.batch_size))

        self.identifier_template = self.config.get('fastq', 'identifier')

        # ensure that path ends in '/'
        self.target_dir = os.path.join(opts.read_path, '')
        path_utils.create_dir_if_not_exist(self.target_dir)

        self.filter_reads = True
        if opts.disable_filtering is True:
            self.filter_reads = False

        self.minqscore_1d = config_utils.find_config_option_else_default(
            self.config, 'pipeline', 'min_qscore_1d', QSCORE_DEFAULT)
        self.minqscore_1dsq = config_utils.find_config_option_else_default(
            self.config, 'pipeline', 'min_qscore_1dsq', QSCORE_DEFAULT)

        self.pass_dir = os.path.join(self.target_dir, 'pass')
        self.fail_dir = os.path.join(self.target_dir, 'fail')

        if self.filter_reads:
            path_utils.create_dir_if_not_exist(self.pass_dir)
            path_utils.create_dir_if_not_exist(self.fail_dir)

    @property
    def basecall_type(self):
        """ Returns the basecall type, as specified in the configuration provided
        to the constructor.
        """
        return self.workflow_type

    def load_read_data(self, fname):
        """ Load the read data from the fast5 file fname.
        
        :param fname: A read fast5 file containing raw data.
        :returns: A tuple containing the scaled raw data and a dict with
            the required metadata.
        """
        return None, None

    def set_read_data(self, read_data):
        """ Set meta-data about a read, which has been loaded from Fast5.
        :param read_data: ReadMetadata object (defined in read_metadata)
        """
        data_id = read_data.data_id
        exp_start_time = read_data.exp_start_time
        offset_seconds = read_data.start_time / read_data.sampling_rate
        self.data[data_id] = read_data

    def output_log(self, data_id, data):
        """ Does nothing. FastQ does not contain log information.
        """

    def output_segmentation(self, data_id, data):
        """ Does nothing. FastQ does not contain segmentation information.
        """
        return None

    def output_basecall_1d(self, data_id, data, seg_group):
        """ Add the Analyses group for the 1D basecall step.
        
        :param data_id: The unique identifier (here: file name) for the read,
            containing format placeholders "{}" for the strand section
            ("template", "complement")
        :param data: A dict containing the results of the 1D basecall
            step of the pipeline analysis.
        :param seg_group: The group name the segmentation results were written to.
        :returns: None.
        """
        has_template = False
        has_complement = False
        sections = []
        if data:
            has_template = ('num_events_template' in data and data['num_events_template'] > 0)
            has_complement = ('num_events_complement' in data and data['num_events_complement'] > 0)
            if has_template:
                sections.append('template')
                # see if the read passes the qscore filter
                passes_filter = False
                if data['mean_qscore_template'] >= self.minqscore_1d:
                    passes_filter = True
                self.data[data_id].add_read_filter_status(passes_filter)
            if has_complement:
                sections.append('complement')

            for section in sections:
                self._add_strand(data_id, data['read_id'], section, data['{}_fastq'.format(section)])
    
    def output_basecall_2d(self, data_id, data, seg_group, bc1d_group):
        """ Add the Analyses group for the 2D basecall step.
        
        :param data_id: The unique identifier for the read.
        :param data: A dict containing the results of the 2D basecall
            step of the pipeline analysis.
        :param seg_group: Unused. [The group name the segmentation results were written to.]
        :param bc1d_group: Unused. [The group name the 1D basecall results were written to.]
        :returns: None.
        """
        if data:
            passes_filter = False
            if data.get('mean_qscore_2d', -1) >= self.minqscore_1dsq:
                passes_filter = True
            self.data[data_id].add_read_filter_status(passes_filter)
            """
            alignment = data.get('2d_alignment_data')
            labls = data.get('2d_labl_data')
            moves = data.get('2d_movement')
            with open('{}.2d_results.txt'.format(read_id), 'w') as fh:
                fh.write('template_index\tcomplement_index\tlabel\tmovement\n')
                last_x = alignment[0][0]
                last_y = alignment[1][0]
                for n in range(len(moves)):
                    x = alignment[0][n]
                    y = alignment[1][n]
                    if x == -1:
                        x = last_x
                    else:
                        last_x = x
                    if y == -1:
                        y = last_y
                    else:
                        last_y = y
                    fh.write('\t'.join(str(val) for val in [x, y, labls[n], moves[n]]) + '\n')
            """
        self._add_strand(data_id, data['read_id'], '2d', data.get('2d_fastq'))

    def output_barcoding(self, data_id, data, seg_group, bc1d_group, bc2d_group):
        """ Stores barcode information.
        """
        barcode_arr = data.get('barcode_arrangement', '')
        if not barcode_arr: return

        if not data_id in self.data:
            raise Exception('File {} has received no strands, but barcoding data.'.format(data_id))
        self.data[data_id].add_barcode_arrangement(barcode_arr)

    def output_alignment(self, data_id, data, seg_group, bc1d_group, bc2d_group):
        """ Stores alignment information
        """
        if data_id not in self.data:
            raise Exception('File {} has received no strands,'
                            ' but alignment data.'.format(data_id))

        for section, section_data in data.items():
            sam_output = section_data.get('alignment_sam_output', '')
            if not self.data[data_id].has_section_data(section):
                raise Exception('File {} section {} has received no strand,'
                                ' but alignment data.'.format(data_id, section))
            self.data[data_id].add_section_sam_output(section, sam_output)

    def output_calib_detection(self, data_id, data, seg_group, bc1d_group, bc2d_group):
        """ Does not store any calibration strand information
        """
        for section, section_data in data.items():
            if len(section_data.get('calibration_strand_sam_output', '')) > 0:
                read_info = self.data[data_id]
                read_info.add_read_calib_status(True)

    def output_filtering(self, data_id):
        """
        Find whether a read passed or failed filtering
        :param data_id: The unique identifier for the read.
        :return: Boolean of whether the read passes or failed filtering
        """
        return self.data[data_id].read_passes_filter()

    def finish_read(self, data_id):
        """ Notify the Data Handler that no more data will come from this read ID.
        This function is unnecessary for Fast5, but needed for FastQ data.
        :param data_id: The unique identifier for the read.
        """
        if data_id in self.data:
            logger = logging.getLogger('albacore')
            read_data = self.data[data_id]
            fastq_string = ''
            sam_header = ''
            sam_string = ''
            sub_dir = self.get_read_subdirectory(read_data, self.filter_reads, self.barcoding)
            batch_counter, in_batch_counter = self._increment_in_batch_counter(sub_dir)
            for section in ['template', 'complement', '2d']:
                if read_data.has_section_data(section):
                    fastq_string += self._build_fastq_string(data_id, section, batch_counter)
                    sam_string += self._build_sam_string(data_id, section, batch_counter)
                    if in_batch_counter == 0:
                        sam_header = self._build_sam_header(data_id, section)
            if fastq_string:
                fastq_filename = self._get_filename(data_id, 'fastq', batch_counter, sub_dir)
                self._write_file(fastq_filename, fastq_string, in_batch_counter)
            if sam_string:
                sam_filename = self._get_filename(data_id, 'sam', batch_counter, sub_dir)
                self._write_file(sam_filename, sam_string, in_batch_counter, sam_header)

            del self.data[data_id]

    ####################################### PRIVATE METHODS #######################################

    def _write_file(self, filename, file_content, in_batch_counter, header=''):
        logger = logging.getLogger('albacore')
        try:
            if in_batch_counter == 0 and os.path.exists(filename):
                logger.warning(
                    'Previously existing file {} will be overwritten.'.format(
                        filename))
            this_path = os.path.dirname(filename)
            if not os.path.exists(this_path):
                os.makedirs(this_path)
            open_type = 'w'
            if (in_batch_counter > 0) or (self.batch_size == 0):
                open_type = 'a'
            with open(filename, open_type) as ofile:
                if header:
                    ofile.write(header + '\n')
                ofile.write(file_content + '\n')
        except Exception as e:
            logger.warning('Could not write to'
                           ' file {}. Error: {}'.format(filename, e))

    def _add_strand(self, data_id, read_id, section, fastq):
        if fastq:
            if data_id not in self.data:
                logger = logging.getLogger('albacore')
                logger.warning('Received data for read {}'
                               ' before meta-data.'.format(data_id))
            else:
                self.data[data_id].add_section_data(section, fastq)

    def _build_identifier(self, data_id, section, batch_counter):
        read_data = self.data[data_id]
        barcode = read_data.barcode_arrangement if self.barcoding else None
        fastq_data = read_data.get_section_data(section)
        section_suffix = ''
        if self.workflow_type is not 'linear':
            section_suffix = {'template': '_t',
                              'complement': '_c',
                              '2d': ''}[section]
        read_id = read_data.read_id
        identifier = self.identifier_template. \
            format(name=fastq_data['name'],
                   section_read_id=read_id + section_suffix,
                   strand_section=section,
                   batch_counter=batch_counter,
                   barcode_id=barcode,
                   **read_data.__dict__  # allows us to change the cfg keywords
                   )
        return identifier

    def _build_fastq_string(self, data_id, section, batch_counter):
        """This is based on the MAQ standard found here:

        http://maq.sourceforge.net/fastq.shtml

        This method produces the "block". Multi-read fastqs are defined as
        blocks appended to one another without any extra carriage returns.

        """
        read_data = self.data[data_id]
        barcode = None
        if self.barcoding:
            barcode = read_data.barcode_arrangement
        fastq_data = read_data.get_section_data(section)
        section_suffix = ''
        if self.workflow_type is not 'linear':
            section_suffix = {'template': '_t',
                              'complement': '_c',
                              '2d': ''}[section]
        read_id = read_data.read_id
        header = self.header_template. \
            format(
            identifier=self._build_identifier(data_id, section, batch_counter),
            name=fastq_data['name'],
            section_read_id=read_id + section_suffix,
            strand_section=section,
            batch_counter=batch_counter,
            barcode_id=barcode,
            **read_data.__dict__  # allows us to change the cfg keywords
            )
        return '@{}\n{}\n+\n{}'.format(header, fastq_data['sequence'],
                                       fastq_data['qstring'])

    def _build_sam_header(self, data_id, section):
        read_data = self.data[data_id]
        try:
            sam_input = read_data.get_section_sam_output(section)
        except ValueError:
            sam_input = ''

        sam_header = []
        # === parse input sam ===
        for sam_line in sam_input.splitlines():
            if sam_line[:3] == '@SQ':
                sam_header += [sam_line]
        return '\n'.join(sam_header)

    def _build_sam_string(self, data_id, section, batch_counter):
        read_data = self.data[data_id]
        try:
            sam_input = read_data.get_section_sam_output(section)
        except ValueError:
            sam_input = ''
        sam_output = []
        # === parse input sam ===
        for sam_line in sam_input.splitlines():
            # (1.) truncate header
            if sam_line[0] == '@': continue
            # (2.) replace identifier template
            id_template = '{identifier}'
            if sam_line.startswith(id_template):
                sam_output += [self._build_identifier(data_id, section,
                                                      batch_counter) + sam_line[
                                                                       len(
                                                                           id_template):]]
            else:
                sam_output += [sam_line]

        return '\n'.join(sam_output)

    def _get_filename(self, data_id, extension, batch_counter, sub_dir):
        """ Determines the name of the current batch file. If self.batch_size == 0,
        the file name is built from informations of the read, otherwise a batch file name
        is produced.

        An empty barcode field is treated as 'unclassified'.
        For each barcode, a separate batch file is used. Each batch file
        is filled with self.batch_size. If this number is exceeded,
        and allow_new_batch is True, a new batch file is started.
        """
        read_data = self.data[data_id]

        # first part of filepath: barcoding folder
        barcode = None
        if self.barcoding:
            barcode = read_data.barcode_arrangement

        filename = self.file_name_template. \
            format(batch_counter=batch_counter,
                   barcode_id=barcode,
                   extension=extension,
                   **read_data.__dict__  # allows us to change the cfg keywords
                   )

        file_path = os.path.join(self.target_dir, sub_dir, filename)

        return file_path
