import fnmatch
from glob import glob
import logging
import numpy as np
import os
import sys
from albacore import log_utils

def list_input_files(opts):
    """ Find available input files, exclude the already processed ones if --resume option is given
        :returns: pair of lists, containing names of files [0] to process, [1] to write and empty summary line for
    """
    logger = logging.getLogger('albacore')
    files = _find_input_files(opts)
    if not opts.resume: return files, []

    prev_finished, prev_failed = _find_previous_output(opts)
    prev_finished_found_cnt = 0
    prev_failed_found_cnt = 0
    remaining = []
    unfinished = []
    
    for file in files:
        basename = os.path.basename(file)
        if basename in prev_finished:
            prev_finished_found_cnt += 1
        elif basename in prev_failed:
            unfinished.append(file)
            prev_failed_found_cnt += 1
        else:
            remaining.append(file)
    prev_finished_remaining_cnt = len(prev_finished) - prev_finished_found_cnt
    prev_failed_remaining_cnt = len(prev_failed) - prev_failed_found_cnt
    if prev_finished_remaining_cnt or prev_failed_remaining_cnt:
        logger.warning('Could not locate {} files(s) from previous run. '
                       'Are your input and save path the same as for the run '
                       'that is to be resumed?'.format(prev_finished_remaining_cnt + prev_failed_remaining_cnt))
    if unfinished:
        logger.info('Previous run started, but could not finish {} reads, '
                    'which will now be added to summary'.format(len(unfinished)))
    if not remaining:
        logger.info('All reads had been processed by previous run. Done.')
    else:
        logger.info('Previous run completed {} reads.'.format(prev_finished_found_cnt))
    logger.info('Will process {} files.'.format(len(remaining)))
    return remaining, unfinished


def _find_input_files(opts):
    """ Traverse the input folder (recursively, if desired) and find all read files.
        :returns: list of absolute paths to all fast5 files in opts.read_path
                  (and sub-directories, if opts.recursive is True)
    """
    logger = logging.getLogger('albacore')

    # Normalize all the paths for safety
    if opts.input:
        opts.input = os.path.abspath(opts.input)
    opts.read_path = os.path.abspath(opts.read_path)
    # Check whether there are any reads to process.
    if opts.input:
        files = []
        if opts.recursive:
            for root, dirnames, filenames in os.walk(opts.input):
                dirnames.sort()
                filenames.sort()
                for filename in fnmatch.filter(filenames, '*.fast5'):
                    files.append(os.path.abspath(os.path.join(root, filename)))
        else:
            files = [os.path.abspath(x) for x in glob(os.path.join(opts.input,
                                                                   '*.fast5'))]
            files.sort()
        if len(files) == 0:
            logger.warning('No read files found in input folder.')
            logger.info('Done.')
        else:
            logger.info('{} files found in input folder'.format(len(files)))
    else:
        print('No option --input_path given, so expecting file list from stdin...')
        files = []
        for file in sys.stdin.readlines():
            files.append(file.strip())
        if len(files) == 0:
            logger.warning('No read files given via stdin.')
            logger.info('Done.')
    return files

def _find_previous_output(opts):
    """ Scan log and summary file of previous Albacore run.
        :returns: pair of sets, containing names of [0] finished and [1] started,
                  but unfinished files 
    """
    logger = logging.getLogger('albacore')
    resume_cancelled = opts.resume == 'cancelled'

    summ = np.genfromtxt(os.path.join(opts.save_path, opts.summfile),
                         dtype=None, delimiter='\t', names=True)
    # if Albacore is invoked with "--resume cancelled", then all unfinished
    # reads should be processed
    logfile = open(os.path.join(opts.save_path, opts.logfile), 'r')
    if resume_cancelled:
        started = set()
    else:
        started = log_utils.find_submitted(logfile.readlines())
    finished = set()
    for filename in summ['filename']:
        filename = filename.decode()
        if not resume_cancelled:
            try:
                started.remove(filename)
            except:
                logger.warning('Filename {} was found in summary, '
                               'but not in log file.'.format(filename))
        finished.add(filename)
    logger.info('Found {} finished reads.'.format(len(finished)))
    return finished, started
