import os
import logging

from albacore import VERSION
from albacore.pipeline_core import get_debug_level


def initialise_logger(save_path, debug=False, logfilename='pipeline.log', resume_mode=None):
    # Initialize the logger.
    logfile = os.path.join(save_path, logfilename)
    logger = logging.getLogger('albacore')
    handler = logging.FileHandler(logfile, 'a' if resume_mode else 'w')
    if debug:
        logger.setLevel(logging.DEBUG)
        handler.setFormatter(logging.Formatter('%(asctime)-15s %(name)s '
                                               '%(levelname)s %(message)s'))
    else:
        logger.setLevel(logging.INFO)
        handler.setFormatter(logging.Formatter('%(asctime)-15s %(message)s'))
    logger.addHandler(handler)
    logger.info(('Resuming after {} pipeline '.format(resume_mode) if resume_mode else '') +
                'ONT Albacore Sequencing Pipeline Software '
                '(version {})'.format(VERSION))
    logger.info('Debug level is {}'.format(get_debug_level()))
    return logger


def log_submission(file_name):
    """ Returns log line that informs about a file being submitted to the pipeline.
        Use this function to ensure that submitted files can be found via find_submitted().
    """
    return 'Submitting file "{}".'.format(file_name)


def find_submitted(log_lines):
    """ Returns list of files whose submission is logged in the given log file.
        :returns: set of strings with all submitted files
    """
    started = set()
    for line in log_lines:
        words = line.strip().split(' ')
        # we expect the logfile to contain lines like
        # 1914-06-28 10:45:38,581 Submitting file "xyz.fast5",
        # produced by log_submission() above.
        if len(words) >= 3 and words[-3] == 'Submitting':
            started.add(words[-1][1:-2])
    return started


def log_open_batch(file_type, batch_num, sub_dir):
    """ Compose a log line that logs opening a batch and matches the format
        expected by find_batches().
    """
    logline = '\tOpening new {} batch\t{}'.format(file_type, batch_num)
    if sub_dir: logline += '\tin folder\t{}'.format(sub_dir)
    return logline


def log_resume_batch(file_type, batch_num, sub_dir):
    """ Compose a log line that logs resuming a batch and matches the format
        expected by find_batches().
    """
    logline = '\tResuming to {} batch\t{}'.format(file_type, batch_num)
    if sub_dir: logline += '\tin folder\t{}'.format(sub_dir)
    return logline


def find_batches(file_type, log_lines):
    """ Iterate through the lines of the log file and build batch counter dictionaries.
        :returns: pair of two dictionaries, as expected for data_handler.current_batch_counter
                  and data_handler.current_in_batch_counter
    """
    current_batch_counter = {}
    current_in_batch_counter = {}
    for line in log_lines:
        words = line.strip().split('\t')
        # we expect the logfile to contain lines like
        # 1914-06-28 10:45:38,581 \tOpening new fast5 batch\t0\tin folder\tfail/
        # produced by log_open_batch() and log_resume_batch above.
        if len(words) > 1 and (words[1] == 'Opening new ' + file_type + ' batch' \
                               or words[1] == 'Resuming to ' + file_type + ' batch'):
            batch_num = int(words[2])
            if len(words) >= 5 and words[3] == 'in folder':
                sub_dir = words[4]
            else:
                sub_dir = ''
            current_batch_counter[sub_dir] = batch_num + 1
            current_in_batch_counter[sub_dir] = -1
    return current_batch_counter, current_in_batch_counter
