from collections import OrderedDict
import logging

barcoding_fields = [('barcode_arrangement', 'unclassified'),
                    ('barcode_score', 0.0),
                    ('barcode_full_arrangement', 'n/a'),
                    ('front_score', float('-infinity')),
                    ('rear_score', float('-infinity')),
                    ('front_begin_index', -1),
                    ('front_foundseq_length', 0),
                    ('rear_end_index', -1),
                    ('rear_foundseq_length', 0),
                    ('kit', 'n/a'),
                    ('variant', 'n/a')]

general_linear_fields = [('filename', 'unknown'),
                         ('read_id', 'unknown'),
                         ('run_id', 'unknown'),
                         ('channel', 0),
                         ('start_time', 0.0),
                         ('duration', 0.0),
                         ('num_events', 0),
                         ('passes_filtering', 'unknown')]


def get_1d_fields(direction):
    return [('{}_start'.format(direction), 0.0),
            ('num_events_{}'.format(direction), 0),
            ('{}_duration'.format(direction), 0.0),
            ('num_called_{}'.format(direction), 0),
            ('sequence_length_{}'.format(direction), 0),
            ('mean_qscore_{}'.format(direction), 0.0),
            ('strand_score_{}'.format(direction), 0.0)]


get_2d_fields = [('sequence_length_2d', 0),
                 ('mean_qscore_2d', 0.0)]

get_1dsq_fields = [('filename1', 'unknown'),
                   ('filename2', 'unknown'),
                   ('read_id1', 'unknown'),
                   ('read_id2', 'unknown'),
                   ('read_id', 'unknown'),
                   ('run_id', 'unknown'),
                   ('channel', 0),
                   ('start_time', 0.0),
                   ('sequence_length_2d', 0),
                   ('mean_qscore_2d', 0.0),
                   ('passes_filtering', 'unknown')]


def get_alignment_fields(basecall_type):
    aligner_empty_summary = [('alignment_genome', 'no_match'),
                             ('alignment_genome_start', -1),
                             ('alignment_genome_end', -1),
                             ('alignment_strand_start', -1),
                             ('alignment_strand_end', -1),
                             ('alignment_num_insertions', -1),
                             ('alignment_num_deletions', -1),
                             ('alignment_num_aligned', -1),
                             ('alignment_num_correct', -1),
                             ('alignment_identity', -1.0),
                             ('alignment_accuracy', -1.0),
                             ('alignment_score', 0.0),
                             ('aligned_speed_bps', 0.0)
                             ]
    alignment_type_map = {'linear': ['template'],
                          'template_complement': ['template', 'complement'],
                          'full_2d': ['2d'],
                          '1dsq': ['2d']}
    alignment_fields = []
    for section in alignment_type_map[basecall_type]:
        alignment_fields.extend([('{}_{}'.format(key, section), val) for (key, val) in aligner_empty_summary])
    return alignment_fields


def get_calib_detection_fields(basecall_type):
    calib_empty_summary = [('calibration_strand_genome', 'no_match'),
                           ('calibration_strand_identity', -1.0),
                           ('calibration_strand_accuracy', -1.0),
                           ('aligned_speed_bps', 0.0)
                           ]
    calib_type_map = {'linear': ['template'],
                      'template_complement': ['template', 'complement'],
                      'full_2d': ['2d'],
                      '1dsq': ['2d']}
    calib_fields = []
    for section in calib_type_map[basecall_type]:
        calib_fields.extend([('{}_{}'.format(key, section), val) for (key, val) in calib_empty_summary])
    return calib_fields


def output_summary(basecall_type, include_calib_detection, include_barcoding, include_alignment):
    summary_fields = []
    summary_fields.extend(general_linear_fields)
    summary_fields.extend(get_1d_fields('template'))
    if basecall_type != 'linear':
        summary_fields.extend(get_1d_fields('complement'))
    if basecall_type == 'full_2d':
        summary_fields.extend(get_2d_fields)
    if include_calib_detection:
        summary_fields.extend(get_calib_detection_fields(basecall_type))
    if include_barcoding:
        summary_fields.extend(barcoding_fields)
    if include_alignment:
        summary_fields.extend(get_alignment_fields(basecall_type))
    # delete duplicates
    summary_fields = OrderedDict((x, True) for x in summary_fields).keys()
    return summary_fields


def output_summary_1dsq(include_calib_detection, include_barcoding, include_alignment):
    summary_fields = get_1dsq_fields
    if include_calib_detection:
        summary_fields.extend(get_calib_detection_fields('1dsq'))
    if include_barcoding:
        summary_fields.extend(barcoding_fields)
    if include_alignment:
        summary_fields.extend(get_alignment_fields('1dsq'))
    return summary_fields


def write_telemetry(telemetry_aggregator, telemetry_output_file):
    """ Write json telemetry from an aggregator to a file.

    :param TelemetryAggregator telemetry_aggregator: aggregator to extract
        telemetry from.
    :param telemetry_output_file: filename to output to.

    """
    logger = logging.getLogger(__name__)
    with open(telemetry_output_file, 'w+') as telemetry_output:
        try:
            telemetry_output.write(telemetry_aggregator.to_json())
            logger.info('Summary telemetry written to '
                        '{}'.format(telemetry_output_file))
        except TypeError:
            logger.error('Error serializing telemetry to json.')
