"""
paired_read_1dsq_data_handler.py
Created on: May 15, 2017
Proprietary and confidential information of Oxford Nanopore Technologies, Limited
All rights reserved; (c)2017: Oxford Nanopore Technologies, Limited
"""
import logging
import os
import shutil
import itertools
import numpy as np
import warnings
from datetime import datetime
from configparser import ConfigParser
from albacore.data_handler import DataHandler
from albacore import __version__ as VERSION
from albacore.fastq_data_handler import FastQDataHandler
from ont_fast5_api.fast5_file import Fast5File
from ont_fast5_api.analysis_tools.basecall_1d import Basecall1DTools
from albacore.read_metadata import ReadMetadata, Read1dsqMetadata


class PairedRead1dSqDataHandler(DataHandler):
    """ A DataHandler subclass for loading data from pairs of Fast5 files and writing
    to FastQ files. Output functions call corresponding functions of FastQDataHandler.
    """

    def __init__(self, config_file, opts, telemetry_aggregator=None):
        """ Constructs the object.

        :param config_file: A configuration file with the basecall configuration
            details used for basecalling.
        :param opts: ConfigParser options from calling the main script
        :param telemetry_aggregator: Optional TelemetryAggregator

        """
        self.config = ConfigParser(interpolation=None)
        self.config.read(config_file)
        self.workflow_type = self.config.get('pipeline', 'basecall_type')
        self.fastq_handler = FastQDataHandler(config_file, opts)
        self.write_fastq = True  # This is forwards compatible with fast5/fastq data handling
        self.path_map = {os.path.basename(full_path): full_path for full_path in
                         opts.file_list}
        self._parse_file_index(opts.file_index, telemetry_aggregator)
        print(len(self.pairs), ' pairs found.')

    def get_pair_ids(self):
        """ Returns a list of ids that can be passed to the load_read_data() method. 
        """
        return list(range(len(self.pairs)))

    @property
    def basecall_type(self):
        """ Returns the basecall type, as specified in the configuration provided
        to the constructor.
        """
        return self.workflow_type

    def load_read_data(self, pair_id):
        """ Load the read data from the fast5 files designated by pair_id.
        
        :param pair_id: A unique identifier for a pair of read files.
        :returns: A tuple containing the a dict of data and a dict of metadata.
        """
        file1, file2 = self.pairs[pair_id]

        meta2 = ReadMetadata(file2)
        metadata = Read1dsqMetadata(file1, meta2, str(pair_id))

        with Fast5File(file1, 'r') as fh:
            group_name = fh.get_latest_analysis('Basecall_1D')
            summary = fh.get_summary_data(group_name)
            stay_prob1 = summary.get('basecall_1d_template', {}).get('stay_prob', 0.0)
            step_prob1 = summary.get('basecall_1d_template', {}).get('step_prob', 0.0)
            skip_prob1 = summary.get('basecall_1d_template', {}).get('skip_prob', 0.0)
            with Basecall1DTools(fh, 'r', group_name) as bc:
                events1 = bc.get_event_data('template')
        with Fast5File(file2, 'r') as fh:
            group_name = fh.get_latest_analysis('Basecall_1D')
            summary = fh.get_summary_data(group_name)
            stay_prob2 = summary.get('basecall_1d_template', {}).get('stay_prob', 0.0)
            step_prob2 = summary.get('basecall_1d_template', {}).get('step_prob', 0.0)
            skip_prob2 = summary.get('basecall_1d_template', {}).get('skip_prob', 0.0)
            with Basecall1DTools(fh, 'r', group_name) as bc:
                events2 = bc.get_event_data('template')
        data = {'template_stay_prob': stay_prob1,
                'template_step_prob': step_prob1,
                'template_skip_prob': skip_prob1,
                'complement_stay_prob': stay_prob2,
                'complement_step_prob': step_prob2,
                'complement_skip_prob': skip_prob2}
        fields = ['mean', 'stdv', 'start', 'length', 'model_state', 'weights']
        for field in fields:
            data['template_' + field] = events1[field]
            data['complement_' + field] = events2[field]
        data['template_weight'] = data.pop('template_weights')
        data['complement_weight'] = data.pop('complement_weights')
        labllen = len(data['template_model_state'][0])
        data['label_len'] = labllen
        all_labels = list(map(lambda x: ''.join(x), itertools.product('ACGT', repeat=labllen)))
        label_map = {label: n for n, label in enumerate(all_labels)}
        template_labels = data.pop('template_model_state')
        complement_labels = data.pop('complement_model_state')
        template_states = [label_map[label] for label in template_labels]
        complement_states = [label_map[label] for label in complement_labels]
        data['template_model_state'] = np.array(template_states, dtype=np.int32)
        data['complement_model_state'] = np.array(complement_states, dtype=np.int32)
        self.fastq_handler.set_read_data(metadata)
        return data, metadata

    def output_segmentation(self, data_id, data):
        """ Add the Analyses group for the segmentation step.
        
        :param data_id: The unique identifier for the read.
        :param data: A dict containing the results of the segmentation
            step of the pipeline analysis.
        :returns: None.
        """
        return None

    def output_basecall_1d(self, data_id, data, seg_group):
        """ Add the Analyses group for the 1D basecall step.
        
        :param data_id: The unique identifier for the read.
        :param data: A dict containing the results of the 1D basecall
            step of the pipeline analysis.
        :param seg_group: The group name the segmentation results were written to.
        :returns: None.
        """
        return None

    def output_basecall_2d(self, data_id, data, seg_group, bc1d_group):
        """ Add the Analyses group for the 2D basecall step.
        
        :param data_id: The unique identifier for the read.
        :param data: A dict containing the results of the 2D basecall
            step of the pipeline analysis.
        :param seg_group: The group name the segmentation results were written to.
        :param bc1d_group: The group name the 1D basecall results were written to.
        :returns: None.
        """
        self.fastq_handler.output_basecall_2d(str(data_id), data, seg_group, bc1d_group)
        return None

    def output_log(self, data_id, data):
        """ Output analysis results from the pipeline.

        :param data_id: The unique identifier for the read.
        :param data: A dict containing at least a key 'log'.
        """
        self.fastq_handler.output_log(data_id, data)

    def output_calib_detection(self, data_id, data, seg_group, bc1d_group, bc2d_group):
        """ Stores calibration strand detection information
        """
        self.fastq_handler.output_calib_detection(data_id, data, seg_group, bc1d_group, bc2d_group)

    def output_barcoding(self, data_id, data, seg_group, bc1d_group, bc2d_group):
        """ Add the Analyses group for the barcoding step.

        :param data_id: The unique identifier for the read.
        :param data: A dict containing the results of the barcoding
            step of the pipeline analysis.
        :param seg_group: The group name the segmentation results were written to.
        :param bc1d_group: The group name the 1D basecall results were written to.
        :param bc2d_group: The group name the 2D basecall results were written to (if any).
        :returns: None.
        """
        return None

    def output_alignment(self, data_id, data, seg_group, bc1d_group, bc2d_group):
        """ Stores alignment information
        """
        if self.write_fastq:
            self.fastq_handler.output_alignment(str(data_id), data, seg_group, bc1d_group, bc2d_group)

    def output_filtering(self, data_id):
        """
        Find whether a read passed or failed filtering
        :param data_id: The unique identifier for the read.
        :return: Boolean of whether the read passes or failed filtering
        """
        return self.fastq_handler.output_filtering(data_id)

    def finish_read(self, data_id):
        """ Notify the Data Handler that no more data will come from this read ID.
        This function is unnecessary for Fast5, but needed for FastQ data.
        :param data_id: The unique identifier for the read.
        """
        self.fastq_handler.finish_read(str(data_id))

    ################################
    #
    # Private methods below.
    #
    ################################

    def _parse_file_index(self, file_index, telemetry_aggregator=None):
        # There is a numpy 1.14 warning raised here that states that we need to
        # "specify the encoding" in order to read unicode strings. If you
        # specify the encoding as the default one ("bytes") then it still raises
        # the VisibleDeprecationWarning, which is not helpful.
        with warnings.catch_warnings():
            try:
                # The warning is only in later versions of numpy. Ugh.
                warnings.simplefilter("ignore", np.VisibleDeprecationWarning)
            except:
                pass
            all_data = np.genfromtxt(file_index, dtype=None, delimiter='\t', names=True)
        print(all_data.size, ' entries found in file index.')

        run_ids = np.unique(all_data['run_id'])
        self.pairs = []
        for run_id in run_ids:
            data = all_data[all_data['run_id'] == run_id]

            # We need to sort this record array by channel and start_time, but this kind of
            # sort uses all fields to resolve ties between the ones we are interested in.
            # Let's create a view of the array that excludes the fields we aren't going to
            # use (which should be most of them).
            keep = ['channel', 'start_time']
            dt = data.dtype
            formats = [dt.fields[name][0] for name in keep]
            offsets = [dt.fields[name][1] for name in keep]
            itemsize = dt.itemsize
            newdt = np.dtype(dict(names=keep, formats=formats, offsets=offsets, itemsize=itemsize))
            table = data.view(newdt)
            table.sort(order=keep)

            # The full array data is now sorted in order of channel first, then start_time.
            filenames = data['filename']
            channels = data['channel']
            starts = data['start_time']
            durations = data['duration']
            seqlens = data['sequence_length_template']
            max_open_pore = self.config.getfloat('prepare_1dsq',
                                                 'max_open_pore')
            max_seq_len_ratio = self.config.getfloat('prepare_1dsq',
                                                     'max_seq_len_ratio')
            for i in range(len(data) - 1):
                if channels[i] != channels[i + 1] or seqlens[i] == 0 or seqlens[i + 1] == 0:
                    # Two consecutive reads must have sequence data and be from
                    # the same channel.
                    continue
                # If we're aggregating telemetry we want info on all the
                # candidates, not just the ones which were successful, so we'll
                # do the telemetry aggregation as long as we're sure we have
                # sensible candidate data.
                if telemetry_aggregator is not None:
                    try:
                        telemetry_aggregator.add_1dsq_candidate_data(data[i],
                                                                     data[i+1],
                                                                     run_id)
                    except KeyError:
                        # We'll have already logged this within telemetry.py
                        pass
                if (starts[i + 1] - starts[i]) > (durations[i] + max_open_pore):
                    # The complement read should be captured very quickly after
                    # the template read.
                    continue
                seqlen_ratio = float(seqlens[i]) / float(seqlens[i + 1])
                if seqlen_ratio > max_seq_len_ratio or seqlen_ratio < 1.0 / max_seq_len_ratio:
                    # Template and complement sequences should be similar lengths.
                    continue
                file1 = self.path_map.get(filenames[i].decode())
                file2 = self.path_map.get(filenames[i + 1].decode())
                if file1 is not None and file2 is not None:
                    self.pairs.append((file1, file2))
