import copy
import os

import albacore.time_utils as time_utils
from ont_fast5_api.fast5_file import Fast5File


class ReadMetadata(object):
    def __init__(self, fname):
        self.filename = fname
        self.barcode_arrangement = 'unclassified'
        self.passed_filter = False
        self.is_calib = False
        self._read_fast5()

    def _read_fast5(self):
        with Fast5File(self.filename, 'r') as fh:
            read_info = fh.status.read_info[0]
            tracking_id = fh.get_tracking_id()
            channel_data = fh.get_channel_info()
            context_tags = fh.get_context_tags()

            self.read_number = fh._get_only_read_number()
            self.raw = fh.get_raw_data(scale=True)

        self.data_id = os.path.basename(self.filename)
        self.read_id = read_info.read_id
        self.start_time = read_info.start_time
        self.channel_id = channel_data['channel_number']
        self.sampling_rate = channel_data['sampling_rate']
        self.run_id = tracking_id['run_id']
        self.label = os.path.splitext(os.path.basename(self.filename))[0]
        self.mux = read_info.start_mux
        self.flowcell_id = tracking_id['flow_cell_id']
        self.device_id = tracking_id['device_id']
        self.hostname = tracking_id['hostname']
        self.exp_start_time = tracking_id['exp_start_time']
        self.median_before = read_info.median_before
        if 'sample_id' not in tracking_id:
            tracking_id['sample_id'] = 'none'
        self.sample_id = tracking_id['sample_id']
        self.tracking_id = tracking_id
        self.context_tags = context_tags
        
        offset_seconds = self.start_time / self.sampling_rate
        start_utc = time_utils.compute_start_timestamp(self.exp_start_time,
                                                       offset_seconds)
        self.start_time_utc = start_utc

        self.section = {}  # for fastq handler to store section data
        self.section_sam_output = {}

    # -------------------------------
    def has_section_data(self, section_name):
        if section_name in self.section:
            return True
        else:
            return False

    def read_passes_filter(self):
        return self.passed_filter

    def read_is_calib(self):
        return self.is_calib

    # -------------------------------
    def replace_filename(self, new_filename):
        self.filename = new_filename

    def add_barcode_arrangement(self, arrangement_data):
        if type(arrangement_data) is not str:
            raise ValueError("Arrangement data should be a string")
        self.barcode_arrangement = arrangement_data

    def add_section_sam_output(self, section_name, sam_output):
        if section_name in self.section_sam_output:
            raise ValueError("Section {} already populated"
                             " with sam output data.".format(section_name))
        self.section_sam_output[section_name] = sam_output

    def add_section_data(self, section_name, section_data):
        if section_name in self.section:
            raise ValueError("Section {} already "
                             "populated; read id {}".format(section_name,
                                                            self.read_id))
        self.section[section_name] = section_data

    def add_read_filter_status(self, passes_filter):
        self.passed_filter = passes_filter

    def add_read_calib_status(self, is_calib):
        self.is_calib = is_calib

    # -------------------------------
    def get_section_data(self, section_name):
        if section_name not in self.section:
            raise ValueError("Read id {} section {}"
                             " has received no strand.".format(self.read_id,
                                                              section_name))
        return self.section[section_name]

    def get_section_sam_output(self, section_name):
        if section_name not in self.section_sam_output:
            raise ValueError("Read id {} has no sam "
                             "output data for section {}".format(self.read_id,
                                                                 section_name))
        return self.section_sam_output[section_name]

    def get_read_data_dict_for_pipeline(self):
        """
        Return a dictionary of the key read metadata for passing in C++
        :return: dict of important read metadata
        """
        read_data_dict = {
            'channel_id': self.channel_id,
            'label': self.label,
            'sampling_rate': self.sampling_rate,
            'start_time': self.start_time,
            'read_id': self.read_id,
            'run_id': self.run_id,
            'data_id': self.data_id
        }
        return read_data_dict

    def get_read_data_dict_for_summary(self):
        """ Get a dictionary of read data for use in summaries.

        :returns: Dictionary of summary read metadata
        :rtype: dict

        """
        results = self.get_read_data_dict_for_pipeline()
        results.update({'median_before': self.median_before})
        return results


class Read1dsqMetadata(ReadMetadata):
    def __init__(self, fname, read_2_metadata, data_id):
        super(Read1dsqMetadata, self).__init__(fname)

        self.read_id1 = self.read_id
        self.read_id2 = read_2_metadata.read_id

        self.label1 = self.label
        self.label2 = read_2_metadata.label

        self.data_id = data_id
        self.read_id = self.read_id + read_2_metadata.read_id

    def get_read_data_dict_for_pipeline(self):
        """
        Return a dictionary of the key read metadata for passing in C++
        :return: dict of important read metadata
        """
        read_data_dict = {
            'channel_id': self.channel_id,
            'label': self.label,
            'label1': self.label1,
            'label2': self.label2,
            'sampling_rate': self.sampling_rate,
            'start_time': self.start_time,
            'read_id': self.read_id,
            'data_id': self.data_id,
            'read_id1': self.read_id1,
            'read_id2': self.read_id2,
            'run_id': self.run_id
        }
        return read_data_dict
