from collections import Counter
import logging
from math import floor
import urllib
from urllib import request


PING_RESPONSE_SUCCESS = b'received'


class Histogram(object):
    """ Stores information for histogram-based telemetry entries """

    def __init__(self, min_bin_val=None, max_bin_val=None, num_bins=None,
                 output_bin_label=None):
        """ Constructs a histogram aggregator.

        Note that max_bin_val is not actually included in the histogram -- the
        range is [min_bin_val, max_bin_val).

        :param min_bin_val: the minimum value for the histogram
        :param max_bin_val: the maximum value for the histogram
        :param num_bins: the number of bins to distribute
        :param output_bin_label: what to call the output bins when converting to
            json

        """
        if num_bins is not None:
            assert(num_bins > 0)
        self.data = Counter()
        self.min_bin_val = min_bin_val
        self.max_bin_val = max_bin_val
        self.num_bins = num_bins
        self.int_bins = False
        if (min_bin_val is not None
                and max_bin_val is not None
                and num_bins is not None):
            self.bin_size = (self.max_bin_val
                             - self.min_bin_val) / self.num_bins
            assert(self.bin_size > 0)
            if float(min_bin_val).is_integer() and self.bin_size.is_integer():
                self.int_bins = True
        self.output_bin_label = output_bin_label

    def __eq__(self, other):
        return (self.num_bins == other.num_bins and
                self.min_bin_val == other.min_bin_val and
                self.max_bin_val == other.max_bin_val and
                self.bin_size == other.bin_size and
                self.output_bin_label == other.output_bin_label and
                self.data == other.data)

    def __ne__(self, other):
        return not self.__eq__(other)

    def increment(self, value):
        """ Finds the correct bin for value and increments its counter.

        :param value: Value which needs its bin incremented

        """
        if (self.min_bin_val is None
                or self.bin_size is None
                or self.num_bins is None):
            raise ValueError('Attmpting to increment a value when histogram '
                             'has not been fully initialized.')
        # If value falls outside of of the bins, put it in the edge bin
        bin_number = min(floor((max(value, self.min_bin_val) - self.min_bin_val)
                               / self.bin_size), self.num_bins - 1)
        bin_value = self.min_bin_val + bin_number * self.bin_size
        if self.int_bins:
            bin_value = int(bin_value)
        self.data[bin_value] += 1

    def combine(self, histogram):
        """ Add data from an existing Histogram into this one.

        Note that no checks are performed to make sure the histograms share the
        same characteristics (min_bin_val, num_bins, etc).

        :param histogram: input Histogram

        """
        assert(isinstance(histogram, Histogram))
        self.data = self.data + histogram.data

    def to_list_of_dicts(self):
        """ Convert the Histogram to a list of dictionaries suitable for output
        in json.

        Each dictionary entry is of the form:

            * <output_bin_label>: bin name
            * "count": bin count

        :returns: list of dictionaries
        :rtype: dict

        """
        output = []
        # We'll make sure there's always at least one element in the histogram
        # This will let us import jsonified histograms later
        if len(self.data) == 0:
            output.append({self.output_bin_label: self.min_bin_val,
                           'count': 0})
        else:
            for key, val in self.data.items():
                if val > 0:
                    output.append({self.output_bin_label: key,
                                   'count': val})
        return output

    def num_entries(self):
        return len(list(self.data))

    @classmethod
    def from_dict(cls, list_of_dicts):
        """ Construct a Histogram from a list of dictionaries, such as you'd
        normally get when loading telemetry from json. Each dictionary entry
        should be of the form

            * <output_bin_label>: <bin_label>
            * count: <count>

        Note that we won't be able to extrapolate what the min and max bin
        values are, or the number of bins, so if this is used with an
        uninitialized Histogram then it won't be possible to increment any of
        the bins later.

        :param list_of_dicts: A list of dictionaries containing histogram data.
        :returns: An instance of Histogram containing the dictionary data
        :rtype: Histogram or None
        :raises: KeyError or IndexError if the list cannot be parsed correctly.

        """
        histogram = cls()
        try:
            first_element = list_of_dicts[0]
            keys = list(first_element.keys())
            output_bin_label = keys[0] if keys[0] != 'count' else keys[1]
            histogram.output_bin_label = output_bin_label
            for entry in list_of_dicts:
                # Don't import anything with a count of zero
                if entry['count'] > 0:
                    histogram.data[entry[output_bin_label]] = entry['count']
        except (KeyError, IndexError):
            raise
        return histogram


def ping(telemetry_aggregator, url):
    logger = logging.getLogger(__name__)
    headers = {'content-type': 'application/json'}
    for entry in telemetry_aggregator.json_entries():
        req = request.Request(url, data=entry.encode('utf-8'),
                              headers=headers)
        try:
            response = request.urlopen(req)
        except urllib.error.URLError as e:
            logger.error('Ping failure: {}'.format(e))
            continue
        result = response.read()
        if result != PING_RESPONSE_SUCCESS:
            logger.warning('Unexpected ping response: {}'.format(result))
