#!/users/bi/lcozzuto/.pyenv/versions/3.6.9/bin/python3.6
"""
full_1dsq_basecaller.py
Created on: May 18, 2017
Proprietary and confidential information of Oxford Nanopore Technologies, Limited
All rights reserved; (c)2016: Oxford Nanopore Technologies, Limited

Main entry point script for running read fast5 files through albacore.

Expected Input:

    *  Folder containing read fast5 files with raw data.
    *  Flow-cell and kit combination.

What It Does:

    * The save path is created if it does not exist.
    * A workspace folder is created under the save path.
    * Read files are copied into the workspace folder.
    * If the recursive option is used, subfolders of the input path are searched
      as well.
    * Read files are processed using the linear basecall workflow, and basecall
      results are written back into the read fast5 files.
      * Files that are completely finished are copied into batch subfolders
        of the workspace folder, maintaining an internal counter and starting
        a new folder when the counter reaches files_per_batch_folder.
      * If the fastq output option is specified, fastq files will also be
        created.
      * Log and summary files are created in the save folder as the program runs.
    * Basecalled read files are processed with the 1d-squared basecaller, with
      pairs of files representing a single double-strand of DNA being inferred
      from the summary file generated by the linear basecall workflow.
    * Fastq files are output containing the 1d-squared basecalls.
"""
import os
import sys
import platform
import subprocess
import pkg_resources
from configparser import ConfigParser
from argparse import ArgumentParser, SUPPRESS

from albacore import VERSION, VERSION_I, MIN_CFG_VERSION_I
from albacore.config_selector import (print_workflows, choose_config,
                                      WORKFLOW_1DSQ)
from albacore.path_utils import initialise_paths, this_extension
from albacore.log_utils import initialise_logger
from albacore.telemetry import generate_analysis_id


def main(opts):
    """ Prepares config values and logging before starting the main workflow
    processing.

    :param opts: Configuration options from the command line.
    :returns: 0 on success.
    :rtype: int

    """
    # Intialise paths and loggers
    opts.read_path = initialise_paths(opts.save_path)
    logger = initialise_logger(opts.save_path, opts.debug, logfilename=opts.logfile)
    if opts.data_path is None:
        opts.data_path = opts.default_path

    # Get the config based on the flowcell, kit and data_path
    # NB opts.config is not an option in full_1dsq so will not exist
    config_path, _ = choose_config(opts.data_path, opts.flowcell, opts.kit)
    if os.path.dirname(config_path) == '':
        config_path = os.path.join(opts.data_path, config_path)
    if not os.path.exists(config_path):
        raise Exception('Config file "{}" does not exist.'.format(config_path))
    config = ConfigParser(interpolation=None)
    config.read(config_path)

    # in case a specific version is specified in the config file, check if this
    # version is in the range [min_config_version, current_version]
    config_version = config.get('pipeline', 'albacore_version', fallback=None)
    if config_version:
        config_version_info = tuple([int(num) for num in
                                     config_version.split('.')])
        if (config_version_info < MIN_CFG_VERSION_I
            or config_version_info > VERSION_I):
            msg = ("Current version of albacore ({}) is incompatible with the "
                   "config file's version ({}).".format(VERSION,
                                                        config_version))
            logger.error(msg)
            raise Exception(msg)

    if not config.has_section('compatibility') or not config.has_option('compatibility', '1dsq_config'):
        raise Exception('The flowcell and kit specified are not compatible with this script.')

    opts.config_1dsq = config.get('compatibility', '1dsq_config')
    if os.path.dirname(opts.config_1dsq) == '':
        opts.config_1dsq = os.path.join(opts.data_path, opts.config_1dsq)
    if not os.path.exists(opts.config_1dsq):
        raise Exception('Config file "{}" does not exist.'.format(opts.config_1dsq))

    # Fix the analysis id so we can associate telemetry produced by the two
    # scripts
    analysis_id = generate_analysis_id()

    # Call the linear workflow.    
    exefile = os.path.join(opts.exe_path, 'read_fast5_basecaller.' +
                           this_extension())
    com = [exefile,
           '-i', opts.input,
           '-t', opts.worker_threads,
           '-s', opts.save_path,
           '-c', config_path,
           '-d', opts.data_path,
           '-n', opts.files_per_batch_folder,
           '-q', opts.reads_per_fastq_batch,
           '-o', opts.output_format,
           '--disable_filtering',
           '--telemetry.analysis_id={}'.format(analysis_id)]
    if this_extension() != 'exe':
        com.insert(0, sys.executable)
    if opts.debug:
        com.extend(['-b'])
    if opts.recursive:
        com.extend(['-r'])
    if opts.align_ref:
        com.extend(['-a', opts.align_ref])
    if opts.resume:
        com.extend(['--resume'])
    if opts.disable_pings:
        com.extend(['--disable_pings'])
    logger.info('Executing command: ' + ' '.join(com))
    res = subprocess.call(com)
    if res != 0:
        logger.error('Call to "read_fast5_basecaller.py" failed. Could not do the linear basecall.')
        return res

    # Call the 1d-squared workflow.
    save_path_1dsq = os.path.join(opts.save_path, '1dsq_analysis')
    exefile = os.path.join(opts.exe_path, 'paired_read_basecaller.' + this_extension())
    file_index = os.path.join(opts.save_path, 'sequencing_summary.txt')
    com = [exefile,
           '-i', opts.read_path,
           '-t', opts.worker_threads,
           '-s', save_path_1dsq,
           '-c', opts.config_1dsq,
           '-d', opts.data_path,
           '-q', opts.reads_per_fastq_batch,
           '-f', file_index,
           '-r',
           '--telemetry.analysis_id={}'.format(analysis_id)]
    if this_extension() != 'exe':
        com.insert(0, sys.executable)
    if opts.barcoding:
        com.extend(['--barcoding'])
    if opts.debug:
        com.extend(['-b'])
    if opts.align_ref:
        com.extend(['-a', opts.align_ref])
    if opts.disable_filtering:
        com.extend(['--disable_filtering'])
    if opts.resume:
        com.extend(['--resume', opts.resume])
    if opts.disable_pings:
        com.extend(['--disable_pings'])
    logger.info('Executing command: ' + ' '.join(com))
    res = subprocess.call(com)
    if res != 0:
        logger.error('Call to "paired_read_basecaller.py" failed. Could not do the 1d-squared basecall.')
        return res

    logger.info('Done.')
    return 0


def this_folder():
    """ We need this in every script as the scripts may not live anywhere near
    the albacore package itself.

    """
    if getattr(sys, 'frozen', False):
        # The application is frozen
        return os.path.dirname(sys.executable)
    else:
        # The application is not frozen
        return os.path.dirname(__file__)


if __name__ == '__main__':
    # Check the data path.
    exe_path = this_folder()
    default_path = os.path.join(exe_path, 'data')
    if not os.path.exists(default_path):
        if pkg_resources.resource_exists('albacore', 'data_versioned'):
            default_path = pkg_resources.resource_filename('albacore',
                                                           'data_versioned')
        elif platform.system() == 'Linux':
            default_path = '/opt/albacore'
        else:
            raise Exception('Package resource "data_versioned" not found and '
                            'no default for platform: {}'.format(platform.system()))

    # Check input parameters.
    list_parser = ArgumentParser(description='ONT Albacore Sequencing Pipeline '
                                             'Software', add_help=False)
    list_parser.add_argument('-l', '--list_workflows', action='store_true',
                             default=False,
                             help='List standard flowcell / kit combinations.')
    list_parser.add_argument('-v', '--version', action='version',
                             help='Print the software version.',
                             version='ONT Albacore Sequencing Pipeline '
                                     'Software (version {})'.format(VERSION), )
    opts, _ = list_parser.parse_known_args()
    if opts.list_workflows:
        print_workflows(default_path, workflow=WORKFLOW_1DSQ)

    parser = ArgumentParser(description='ONT Albacore Sequencing Pipeline '
                                        'Software', parents=[list_parser])
    parser.add_argument('-i', '--input', required=True,
                        help='Folder containing read fast5 files.')
    parser.add_argument('-t', '--worker_threads', required=True,
                        help='Number of worker threads to use.')
    parser.add_argument('-s', '--save_path', required=True,
                        help='Path to save output.')
    parser.add_argument('--resume', metavar='REASON', nargs='?', const='failed', default='',
                        help='Resume previous run for the given save path. '
                             'Optional parameter X is for debugging purposes only.')
    parser.add_argument('--barcoding', required=False, default=False,
                        action='store_true', help='Search for barcodes to '
                                                  'demultiplex sequencing data.')
    parser.add_argument('-f', '--flowcell', required=True,
                        help='Flowcell used during the sequencing run.')
    parser.add_argument('-k', '--kit', required=True,
                        help='Kit used during the sequencing run.')
    parser.add_argument('-d', '--data_path', required=False, default=None,
                        help='Optional path to model files.')
    parser.add_argument('-b', '--debug', action='store_true', default=False,
                        help='Output additional debug information to the log.')
    # 'name of fasta genome to align against (will call bwa index if required).'
    parser.add_argument('-a', '--align_ref', required=False, default=None, help=SUPPRESS)
    parser.add_argument('-r', '--recursive', action='store_true', default=False,
                        help='Recurse through subfolders for input data files.')
    parser.add_argument('-n', '--files_per_batch_folder', required=False, default='4000',
                        help='Maximum number of files in each batch subfolder. '
                             'Set to 0 to disable batch subfolders.')
    parser.add_argument('-o', '--output_format', required=False, default='fastq,fast5',
                        help='desired output format, can be fastq,fast5 or fast5. '
                             'Applies only to 1d part of process')
    parser.add_argument('-q', '--reads_per_fastq_batch', required=False, default='4000',
                        help='number of reads per FastQ batch file. Set to 0 to receive '
                             'one reads per file and file names which include the read ID.')
    parser.add_argument('--disable_filtering', action='store_true',
                        default=False,
                        help='Disable filtering into pass/fail folders')
    parser.add_argument('--disable_pings', action='store_true', default=False,
                        help='Do not send summary information about the run')

    opts, unknown_args = parser.parse_known_args()
    if unknown_args:
        print()
        print('    Unknown arguments: {}'.format(' '.join(unknown_args)))
        print('    Config overrides are not supported for '
              'full_1dsq_basecaller.')
        print('    Call the individual 1d and 1dsq basecall scripts to use '
              'this feature.')
        print()
        sys.exit(1)
    opts.logfile = 'full_1dsq_pipeline.log'

    if opts.output_format == 'fastq':
        opts.output_format = 'fastq,fast5'
    opts.default_path = default_path
    opts.exe_path = exe_path
    sys.exit(main(opts))
