#!/users/bi/lcozzuto/.pyenv/versions/3.6.9/bin/python3.6
"""
paired_read_basecaller.py
Created on: May 17, 2017
Proprietary and confidential information of Oxford Nanopore Technologies, Limited
All rights reserved; (c)2016: Oxford Nanopore Technologies, Limited

Main entry point script for running 1d-squared basecalling with albacore.

Expected Input:

    *  Folder containing read fast5 files which have been 1d basecalled.
    *  Summary file from 1d basecalling of fast5 data.
    *  Configuration file to use. This will contain analysis parameters.
    *  Path to save log and intermediate files.

What It Does:

    * The save path is created if it does not exist.
    * A workspace folder is created under the save path.
    * Read files are scanned in input folder.
    * If the recursive option is used, subfolders of the input path are searched
      as well.
    * Read files are processed, and basecall results are written to fastq files
      in the workspace folder.
    * Log and summary files are created in the save folder as the program runs.

Notes:
    The --config parameter can be the full path to the config file, or just
    the filename of one of the standard config files provided with the
    software. If no path is provided, it will first look for the 'data' folder
    within the same path as this script. This is appropriate for checked out
    copies of the code. If this path does not exist, it assumed you are running
    an installed version of the code, and looks in the default installation
    path of the project, which is platform specific.

    The configuration file will specify model files by filename only. If you
    want to use a model that is not in the standard data folder, you will need
    to specify the path with the --data_path option. Note that ALL models used
    must be in the specified folder.

    This script spawns multiple threads, configured with the --worker_threads
    option. However, all IO and some of the work for aggregating basecall
    results is done in the master thread. So for processing large amounts of
    data on a system with lots of cores (ie more than 8), it is recommended
    that you distribute the data into multiple folders, and run multiple
    instances of this script, each with its own input folder.
"""
import os
import sys
import time
import fnmatch
import logging
import platform
from configparser import ConfigParser

import pkg_resources
from glob import glob
from argparse import ArgumentParser, SUPPRESS

from albacore import VERSION, VERSION_I, MIN_CFG_VERSION_I
from albacore.log_utils import initialise_logger
from albacore.path_utils import initialise_paths
from albacore.pipeline import Pipeline
from albacore.summary import SummaryCSV
from albacore.paired_read_1dsq_data_handler import PairedRead1dSqDataHandler

from albacore.config_utils import (get_barcoding_options, parse_config_override,
                                   resume_options_2d, parse_telemetry_options)
from albacore.config_selector import print_workflows, WORKFLOW_1DSQ
from albacore.output_utils import output_summary_1dsq, write_telemetry
from albacore.input_utils import list_input_files

from albacore.aligner_utils import initialise_aligner
from albacore.utilities.pipeline_utils import get_progress_bar
from albacore.telemetry import TelemetryAggregator
from albacore import telemetry_utils


def main(opts, config_override):
    """ Prepares config values and logging before starting the main workflow
    processing.

    :param opts: Configuration options from the command line.
    :param config_override: dict whose key-value pair override the
        corresponding values from the configuration file.
    :returns: 0 on success.
    :rtype: int

    """
    # Intialise paths and loggers
    opts.read_path = initialise_paths(opts.save_path)
    logger = initialise_logger(opts.save_path, opts.debug, logfilename=opts.logfile, resume_mode=opts.resume)
    if opts.data_path is None:
        opts.data_path = opts.default_path

    if os.path.dirname(opts.config) == '':
        opts.config = os.path.join(opts.data_path, opts.config)
    if not os.path.exists(opts.config):
        raise Exception('Config file "{}" does not exist.'.format(opts.config))
    config = ConfigParser(interpolation=None)
    config.read(opts.config)

    # Clear out any compatability information in the config file
    config.remove_section('compatibility')

    if config.has_section('basecaller_2d'):
        config.set('basecaller_2d', 'model_path', opts.data_path)

    # in case a specific version is specified in the config file, check if this
    # version is in the range [min_config_version, current_version]
    config_version = config.get('pipeline', 'albacore_version', fallback=None)
    if config_version:
        config_version_info = tuple([int(num) for num in
                                     config_version.split('.')])
        if (config_version_info < MIN_CFG_VERSION_I
            or config_version_info > VERSION_I):
            msg = ("Current version of albacore ({}) is incompatible with the "
                   "config file's version ({}).".format(VERSION,
                                                        config_version))
            logger.error(msg)
            raise Exception(msg)
        config.remove_option('pipeline', 'albacore_version')

    # Add barcoding options if required (after looking override)
    if ('barcode_detector', 'config_file') in config_override:
        config.set('barcode_detector', 'config_file', config_override[('barcode_detector', 'config_file')])
    get_barcoding_options(config, opts.barcoding, opts.data_path, logger)

    # add override options
    for key, value in config_override.items():
        (section, parameter) = key
        if not config.has_section(section):
            config.add_section(section)
        config.set(section, parameter, value)
        logger.info('config parameter {}.{} overwritten by command-line value {}'.format(section,parameter, value))

    # find external aligners and references; delete respective config section
    # if alignment or calibration strand detection is disabled
    references_loaded = \
        initialise_aligner(config, opts.align_ref, opts.data_path, 'aligner') + \
        initialise_aligner(config, None, opts.data_path, 'calib_detector')

    desc_file = os.path.join(opts.data_path, 'layout_basecall_1dsq.jsn')
    config.set('pipeline', 'desc_file', desc_file)
    config.set('pipeline', 'log_intermodule_messages', str(opts.debug))

    files, additional_summary_lines = list_input_files(opts)

    # Initialize the summary file handler.
    include_calib_detection = config.has_section('calib_detector')
    include_alignment = config.has_option('aligner', 'reference')
    include_barcoding = config.has_option('barcode_detector',
                                          'arrangements_files')
    logger.info('include calibration strand detection: {}'.format(include_calib_detection))
    logger.info('include alignment: {}'.format(include_alignment))
    logger.info('include barcoding: {}'.format(include_barcoding))

    summary_fields = output_summary_1dsq(include_calib_detection,
                                         include_barcoding, include_alignment)
    telemetry_enabled = not opts.disable_pings
    (telemetry_urls,
     telemetry_segment_duration,
     telemetry_analysis_id) = parse_telemetry_options(config)
    telemetry_output_file = os.path.join(opts.save_path,
                                         'sequencing_telemetry_1dsq.js')
    telemetry_aggregator = TelemetryAggregator(
        telemetry_segment_duration,
        albacore_opts=opts,
        include_1d_basecalling=False,
        include_1dsq_basecalling=True,
        include_calibration=include_calib_detection,
        include_alignment=include_alignment,
        include_barcoding=include_barcoding,
        analysis_id=telemetry_analysis_id)
    logger.info('telemetry enabled: {}'.format(telemetry_enabled))
    try:
        with SummaryCSV(os.path.join(opts.save_path, opts.summfile),
                        summary_fields, start_new_file=not opts.resume) as summary:
            for filename in additional_summary_lines:
                summary.write({'filename': os.path.basename(filename)})
            process_pipeline(opts, config, files, summary, telemetry_aggregator)
        logger.info('Done.')
    except Exception as exc:
        logger.exception('Fatal error in pipeline')
        print('Fatal error in pipeline:', exc, file=sys.stderr)
    logger.info('Writing telemetry')
    write_telemetry(telemetry_aggregator, telemetry_output_file)
    if telemetry_enabled:
        for url in telemetry_urls:
            logger.info('Pinging telemetry to {}'.format(url))
            telemetry_utils.ping(telemetry_aggregator, url)
    if references_loaded:
        unload_command = config.get('aligner', 'unload_command'). \
            format(executable=config.get('aligner', 'executable'),
                   reference=config.get('aligner', 'reference')
                  )
        print('\n    CAVE: Reference is still loaded. To unload, call:\n'
              '    ' + unload_command + '\n')
    return 0


def process_pipeline(opts, config, files, summary, telemetry_aggregator=None):
    logger = logging.getLogger('albacore')
    num_workers = opts.worker_threads if opts.worker_threads > 0 else 1
    reverse_direction = config.getboolean('basecaller', 'reverse_direction',
                                          fallback=False)
    u_substitution = config.getboolean('basecaller', 'u_substitution',
                                       fallback=False)
    new_config = os.path.join(opts.save_path, opts.cfgfile)
    with open(new_config, 'w') as cfg_out:
        config.write(cfg_out)

    # Initialize the basecall pipeline.
    opts.file_list = files
    opts.barcoding = False
    data_handler = PairedRead1dSqDataHandler(new_config, opts)
    pipeline = Pipeline(opts.worker_threads, data_handler, new_config,
                        reverse_direction=reverse_direction,
                        u_substitution=u_substitution,
                        telemetry_aggregator=telemetry_aggregator)

    pair_ids = data_handler.get_pair_ids()
    num_pairs = len(pair_ids)
    pair_index = 0
    pairs_complete = 0

    pbar = get_progress_bar(opts.debug, num_pairs)

    no_more_pairs = False
    keep_going = True
    while keep_going:
        num_cached_reads = pipeline.cache_completed_calls()
        free_workers = pipeline.workers_ready()
        if free_workers == num_workers and no_more_pairs:
            keep_going = False
        will_sleep = True

        if free_workers > 0 and pair_index < num_pairs:
            # Submit new jobs before processing complete ones, so that our worker
            # threads are not left idle.
            submit = min(free_workers, num_pairs - pair_index)
            for i in range(submit):
                short_name1, short_name2 = (os.path.basename(data_handler.pairs[pair_index][0]),
                                            os.path.basename(data_handler.pairs[pair_index][1]))
                logger.info('Submitting files "{}" and "{}".'.format(short_name1, short_name2))
                try:
                    status = pipeline.submit_read(pair_ids[pair_index])
                    if not status:
                        logger.warning('Could not extract read data from '
                                       'files "{}" and "{}".'.format(short_name1, short_name2))
                except Exception:
                    logger.exception('Could not submit '
                                     'files "{}" and "{}".'.format(short_name1, short_name2))
                pair_index += 1
            will_sleep = False

        if pair_index == num_pairs:
            no_more_pairs = True

        if num_cached_reads > 0:
            # Process the data from the completed workers.
            summary_data = pipeline.process_cached_calls()
            for line in summary_data:
                line['filename1'] = '{}.fast5'.format(line['label1'])
                line['filename2'] = '{}.fast5'.format(line['label2'])
                pairs_complete += 1
                pbar.update(pairs_complete)
                if 'error_message' in line:
                    logger.warning('Error processing '
                                   'files "{}" and "{}".'.format(line['filename1'], line['filename2']))
                    logger.warning(line['error_message'])
                else:
                    summary.write(line)
                    logger.info('Finished processing '
                                'files "{}" and "{}".'.format(line['filename1'], line['filename2']))
            will_sleep = False

        if will_sleep:
            time.sleep(0.01)

    pbar.finish()


def this_folder():
    """ We need this in every script as the scripts may not live anywhere near
    the albacore package itself.

    """
    if getattr(sys, 'frozen', False):
        # The application is frozen
        return os.path.dirname(sys.executable)
    else:
        # The application is not frozen
        return os.path.dirname(__file__)


if __name__ == '__main__':
    # Check the data path.
    default_path = os.path.join(this_folder(), 'data')
    if not os.path.exists(default_path):
        if pkg_resources.resource_exists('albacore', 'data_versioned'):
            default_path = pkg_resources.resource_filename('albacore',
                                                           'data_versioned')
        elif platform.system() == 'Linux':
            default_path = '/opt/albacore'
        else:
            raise Exception('Unsupported platform.')

    # Check input parameters.
    list_parser = ArgumentParser(description='ONT Albacore 1D-Squared Sequencing Pipeline '
                                             'Software', add_help=False)
    list_parser.add_argument('-l', '--list_workflows', action='store_true',
                             default=False,
                             help='List standard flowcell / kit combinations.')
    list_parser.add_argument('-v', '--version', action='version',
                             help='Print the software version.',
                             version='ONT Albacore 1D-Squared Sequencing Pipeline '
                                     'Software (version {})'.format(VERSION), )

    opts, _ = list_parser.parse_known_args()

    if opts.list_workflows:
        print_workflows(default_path, WORKFLOW_1DSQ)

    parser = ArgumentParser(description='ONT Albacore 1D-Squared Sequencing Pipeline '
                                        'Software', parents=[list_parser])
    parser.add_argument('-i', '--input', required=False, default=None,
                        help='Folder containing read fast5 files '
                             '(if not present, will expect file names on stdin).')
    parser.add_argument('-f', '--file_index', required=True,
                        help='Summary file for indexing potential paired reads.')
    parser.add_argument('-t', '--worker_threads', required=True, type=int,
                        help='Number of worker threads to use.')
    parser.add_argument('-s', '--save_path', required=True,
                        help='Path to save output.')
    parser.add_argument('--resume', metavar='REASON', nargs='?', const='failed', default='',
                        help='Resume previous run for the given save path. '
                             'Optional parameter X is for debugging purposes only.')
    parser.add_argument('--barcoding', required=False, default=False,
                        action='store_true', help='Search for barcodes to '
                                                  'demultiplex sequencing data.')
    # 'name of fasta genome to align against (will call bwa index if required).'
    parser.add_argument('-a', '--align_ref', required=False, default=None, help=SUPPRESS)
    parser.add_argument('-c', '--config', required=True, default=None,
                        help='Configuration file to use.')
    parser.add_argument('-d', '--data_path', required=False, default=None,
                        help='Optional path to model files.')
    parser.add_argument('-b', '--debug', action='store_true', default=False,
                        help='Output additional debug information to the log.')
    parser.add_argument('-r', '--recursive', action='store_true', default=False,
                        help='Recurse through subfolders for input data files.')
    parser.add_argument('-q', '--reads_per_fastq_batch', required=False,
                        type=int, default=4000,
                        help='number of reads per FastQ batch file. Set to 0 to receive '
                             'one reads per file and file names which include the read ID.')
    parser.add_argument('--disable_filtering', action='store_true',
                        default=False,
                        help='Disable filtering into pass/fail folders')
    parser.add_argument('--disable_pings', action='store_true', default=False,
                        help='Do not send summary information about the run')

    opts, unknown_args = parser.parse_known_args()
    opts.logfile = 'pipeline.log'
    opts.summfile = 'sequencing_1dsq_summary.txt'
    opts.cfgfile = 'configuration.cfg'
    if opts.resume:
        resume_options_1d(opts, unknown_args) # if no log file or no summary file is found,
                                              # opts.resume will be reset to False
                                              # If config, log and summary file are found, unknown_args
                                              # will be cleared

    opts.default_path = default_path
    config_override = parse_config_override(unknown_args)

    sys.exit(main(opts, config_override))
